"""
Cogniac API Python SDK

Copyright (C) 2016 Cogniac Corporation.

This client library provides access to most of the common functionality of the Cogniac public API.


CogniacConnection(username=None,
                  password=None,
                  api_key=None,
                  tenant_id=None,
                  timeout=60,
                  url_prefix="https://api.cogniac.io/1")

        Create an authenticated CogniacConnection with the following credentials:


        username (String):            The Cogniac account username (usually an email address).
                                      If username is None, then use the contents of the
                                      COG_USER environment variable as the username.

        password (String):            The associated Cogniac account password.
                                      The password can also be supplied via the COG_PASS
                                      environment variable.

        api_key (String):             A Cogniac-issued API key that can be used as a substitute for
                                      a username+password.  The api_key can also be supplied via the
                                      COG_API_KEY environment variable.

        tenant_id (String):           Cogniac tenant_id with which to assume credentials.
                                      This is only required if the user is a member of multiple tenants.
                                      If tenant_id is None, and the user is a member of multiple tenants
                                      then use the contents of the COG_TENANT environment variable
                                      will be used as the tenant.

        url_prefix (String):          Cogniac API url prefix.
                                      Defaults to "https://api.cogniac.io/1" for the Cogniac cloud system.
                                      If you are accessing an 'on-prem' version of the Cogniac system,
                                      please set this accordingly (e.g. 'https://your_company_name.local.cogniac.io/1'
                                      or a custom DNS prefix assigned by your internal IT.)
                                      The url_prefix can alternatively be set via the COG_URL_PREFIX environment variable.

        If a user is a member of multiple tenants the user can retrieve his list of associated
        tenants via the CogniacConnection.get_all_authorized_tenants() classmethod.


CogniacConnection has a number of helper functions for working with Cogniac
common Cogniac objects such as applications, subjects, and media:

    get_all_applications
        return CogniacApplications for all applications belonging to the currently authenticated tenant

    get_application
        return an existing CogniacApplication

    create_application
        Create a new CogniacApplication

    get_all_subjects
        return CogniacSubjects for all subjects belonging to the currently authenticated tenant

    get_subject
        return an existing CogniacSubject

    create_subject
        Create a CogniacSubject

    get_media
        return CogniacMedia object for an existing media item

    create_media
        create a new cogniac media item

    get_tenant
        return the currently authenticated CogniacTenant


CogniacEdgeFlow(url_prefix, timeout=60)
    An object representing an EdgeFlow in the Cogniac System.

    An instance of this class optionally provides a client interface that 
    communicates directly with a local EdgeFlow's APIs. The local EdgeFlow APIs
    are useful for application such as uploading images to the EdgeFlow for 
    fast processing "on the edge" without first uploading the images to
    CloudCore.

    url_prefix (String):          Cogniac API url prefix.
                                    The url_prefix can alternatively be set via the COG_GW_URL_PREFIX environment variable.


CogniacApplication
    An object representing an Application in the Cogniac System.

    Applications are the main locus of activity within the Cogniac System.

    This class manages applications within the Cogniac System via the
    Cogniac public API application endpoints.

    Create a new application with
    CogniacConnection.create_application() or CogniacApplication.create()

    Get an existing application with
    CogniacConnection.get_application() or CogniacApplication.get()

    Get all tenant's applications with
    CogniacConnection.get_all_applications() or CogniacApplication.get_all()

    Writes to mutable CogniacApplication attributes are saved immediately via the Cogniac API.


 CogniacSubject
    An object representing a Subject in the Cogniac System.

    Subjects are a central organizational mechanism in the Cogniac system.
    A subject is any user-defined concept that is relevant to images or video.
    More generally a subject can represent any logical grouping of images of video.

    Most Cogniac applications work by taking input media from user-defined subjects
    and outputing those media to other user-defined subjects based on the content
    of the media.

    Create a new subject with
    CogniacConnection.create_subject() or CogniacSubject.create()

    Get an existing subject with
    CogniacConnection.get_subject() or CogniacSubject.get()

    Get all tenant's subject with
    CogniacConnection.get_all_subjects() or CogniacSubject.get_all()

    Writes to mutable CogniacSubjects attributes are saved immediately via the Cogniac API.

CogniacMedia
    CogniacMedia objects contain metadata for media files that has been input into the Cogniac System.
    New CogniacMedia can be created by specifying a local filename containing a still image or video.
    Existing CogniacMedia can be retrieved by media_id.

CogniacTenant
    An object representing a Tenant in the Cogniac System

"""

from .cogniac import CogniacConnection
from .edgeflow import CogniacEdgeFlow
from .app import CogniacApplication
from .tenant import CogniacTenant
from .subject import CogniacSubject
from .media import CogniacMedia
from .common import CredentialError, ServerError, ClientError
from .user import CogniacUser
from .external_results import CogniacExternalResult
from .ops_review import CogniacOpsReview
from .network_camera import CogniacNetworkCamera



"""
CogniacEdgeFlow Object Client

Copyright (C) 2016 Cogniac Corporation
"""

import os
import re
from time import time
import six
import requests
from retrying import retry
from requests.packages.urllib3 import Retry
from requests.adapters import HTTPAdapter

from .common import server_error, raise_errors
from .media import file_creation_time


IP_REGEX = re.compile('^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$')


CLOUDFLOW_PREFIX = "cloudflow.cogniac.io"


@six.python_2_unicode_compatible
class CogniacEdgeFlow(object):
    """
    CogniacEdgeFlow

    connnection (CogniacConnection):     Authenticated CogniacConnection object.

                                         If unspecified, CloudCore functions will not be available.

    url_prefix (String):                 Cogniac EdgeFlow API url prefix.

                                         This is a local URL of a physical EdgeFlow.

                                         Defaults to `None`. If unspecified, the EdgeFlow's local APIs will not be available

    Connect to a Cogniac EdgeFlow and maintain session state.

    Class definition for an object that stores information about a physical
    Cogniac EdgeFlow and methods that provide a client
    interface for programmatically managing and requesting work to be done
    on the EdgeFlow.

    A Cogniac EdgeFlow object's methods can be used to trigger media capture
    (e.g., triggering cameras to save images to an EdgeFlow) and ingesting
    media from another host on the same network.
    """

    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def get(cls, connection, edgeflow_id):
        """
        Get a single EdgeFlow.
        connnection (CogniacConnection): Authenticated CogniacConnection object
        edgeflow_id (String): the unique identifier of a EdgeFlow object
        returns CogniacEdgeFlow object
        """
        edgeflow_dict = None
        resp = connection._get("/1/gateways/{}".format(edgeflow_id))
        return CogniacEdgeFlow(connection, resp.json())

    @classmethod
    def get_all(cls, connection):
        """
        return all CogniacEdgeFlow objects belonging to the currently authenticated tenant

        connnection (CogniacConnection):     Authenticated CogniacConnection object
        """
        resp = connection._get('/1/tenants/%s/gateways' % connection.tenant.tenant_id)
        edgeflows = resp.json()['data']
        return [CogniacEdgeFlow(connection, edgeflow) for edgeflow in edgeflows]

    def __init__(self, connection, edgeflow_dict, timeout=60):
        """
        Initialize a CogniacEdgeFlow object.
        """
        if not edgeflow_dict:
            edgeflow_dict = {}
        super(CogniacEdgeFlow, self).__setattr__('_edgeflow_keys', edgeflow_dict.keys())

        self._cc = connection
        self._edgeflow_keys = edgeflow_dict.keys()
        for k, v in edgeflow_dict.items():
            super(CogniacEdgeFlow, self).__setattr__(k, v)

        self.timeout = timeout

        self.__initialize()

    def __setattr__(self, name, value):
        if name not in self._edgeflow_keys:
            super(CogniacEdgeFlow, self).__setattr__(name, value)
            return
        data = {name: value}
        resp = self._cc._post("/1/gateways/%s" % self.gateway_id, json=data)
        for k, v in resp.json().items():
            super(CogniacEdgeFlow, self).__setattr__(k, v)

    def __str__(self):
        return "%s (%s)" % (self.name, self.gateway_id)

    def __repr__(self):
        return self.__str__()

    # -------------------------------------------------------------------------
    #
    #  EdgeFlow API.
    #
    # -------------------------------------------------------------------------
    def __set_url_prefix(self):
        """ Setup url prefix based on model of Edgeflow (cloudflow vs edgeflow).
            For local http actions, the WAN0 ip address is used"""
        self.url_prefix = None
        if 'cf' in self.model.lower():
            # cloudflow uses https based post request with token
            self.url_prefix = 'https://{}.{}'.format(self.gateway_id, CLOUDFLOW_PREFIX)
            self._post = self._cc._post
        else:
            # use wan0 if available as the http destination ip addres
            try:
                ifconfigs = self.status(subsystem_name='ifconfig', limit=1).next()
                self.ip_address = ifconfigs['status']['wan0']['ip']
                if IP_REGEX.match(self.ip_address):
                    self.url_prefix = 'http://{}:8000'.format(self.ip_address)
            except:
                raise

    def __initialize(self):
        self.session = requests.Session()

        self.__set_url_prefix()

        # Configure session with appropriate retries.
        self.session.mount('https://', HTTPAdapter(
            max_retries=Retry(connect=5,
                              read=5,
                              status=5,
                              redirect=2,
                              backoff_factor=.001,
                              status_forcelist=(500, 502, 503, 504))))

    @retry(stop_max_attempt_number=3, retry_on_exception=lambda e: isinstance(e, Exception))
    def _get(self, url, timeout=None, **kwargs):
        """
        Wrapper method to retry an HTTP GET request if an Exception is raised
        when making the request.
        """
        if not url.startswith("http"):
            url = self.url_prefix + url
        if timeout is None:
            timeout = self.timeout
        resp = self.session.get(url, timeout=timeout, **kwargs)
        raise_errors(resp)
        return resp

    @retry(stop_max_attempt_number=3, retry_on_exception=lambda e: isinstance(e, Exception))
    def _post(self, url, timeout=None, **kwargs):
        """
        Wrapper method to retry an HTTP POST request if an Exception is raised
        when making the request.
        """
        if not url.startswith("http"):
            url = self.url_prefix + url
        if timeout is None:
            timeout = self.timeout
        resp = self.session.post(url, timeout=timeout, **kwargs)
        raise_errors(resp)
        return resp

    def get_version(self):
        resp = self._get("/1/version")
        return resp.json()


    def process_media(self,
                      subject_uid,
                      filename,
                      external_media_id=None,
                      media_timestamp=None,
                      domain_unit=None,
                      post_url=None):
        """
        Uploads a media file object to an EdgeFlow device.

        subject_uid                               A subject's unique identifier.
        filename (str):                   Local filename or http/s URL of image or video media file
        external_media_id (str):          Optional arbitrary external id for this media
        media_timestamp (float):          Optional actual timestamp of media creation/occurance time
        domain_unit (str):                Optional domain id (e.g. serial number) for set assignment grouping.
                                          Media with the same domain_unit will always be assigned to the same
                                          training or validation set. Set this to avoid overfitting when you have
                                          multiple images of the same thing or almost the same thing.
        post_url (str):                   Optional URL where media detections can be posted.

        Returns a MediaDetections list.
        """

        args = dict()
        if external_media_id is not None:
            args['external_media_id'] = external_media_id
        if media_timestamp is not None:
            args['media_timestamp'] = media_timestamp
        if domain_unit is not None:
            args['domain_unit'] = domain_unit
        if post_url is not None:
            args['post_url'] = post_url

        if filename.startswith('http'):
            raise Exception("The media file must be uploaded from local storage.")
        else:  # local filename
            fstat = os.stat(filename)
            if 'media_timestamp' not in args:
                # set the unspecified media timestamp to the earliest file time we have
                args['media_timestamp'] = file_creation_time(filename)

        @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
        def upload():
            if filename.startswith('http'):
                raise Exception("The media file must be uploaded from local storage.")
            else:
                files = {'file': open(filename, 'rb')}
            url = "%s/%s/%s" % (self.url_prefix, "1/process", subject_uid)
            resp = self._post(url, data=args, files=files)
            return resp

        resp = upload()
        return resp.json()

    # -------------------------------------------------------------------------
    #
    #  CloudCore API.
    #
    # -------------------------------------------------------------------------

    def time_bound_media_upload(self, start_time, end_time):
        """
        Request a time region to upload iamges from the EdgeFlow to CloudCore
        Must specify start_time and stop_time in unix time
        """
        args = dict()
        args['start_time'] = start_time
        args['end_time'] = end_time
        self._cc._post("/1/gateways/%s/event/time_bound_media_upload" % (self.gateway_id), json=args)


    def flush_upload_queue(self, start_time=None, end_time=None):
        """
        Flush the EdgeFlow to Core upload queue
        """
        args = dict()
        if start_time is not None:
            args['start_time'] = start_time
        if end_time is not None:
            args['end_time'] = end_time
        self._cc._post("/1/gateways/%s/event/flush_upload_queue" % (self.gateway_id), json=args)


    def factory_reset(self):
        """
        Factory reset the EdgeFlow.
        This results in deletion of this EdgeFlow object in CloudCore/SiteCore!
        """
        self._cc._post("/1/gateways/%s/event/factory_reset" % (self.gateway_id))

    def upgrade(self, software_version):
        """
        Upgrade the edgeflow to the specified software version
        """
        args = dict()
        if software_version is not None:
            args['software_version'] = software_version
        self._cc._post("/1/gateways/%s/event/upgrade" % (self.gateway_id), json=args)

    def set_boot_software_version(self, software_version):
        """
        Upgrade the edgeflow to the specified software version
        """
        args = dict()
        if software_version is not None:
            args['software_version'] = software_version
        self._cc._post("/1/gateways/%s/event/set_boot_software_version" % (self.gateway_id), json=args)

    def reboot(self):
        """
        Reboot the EdgeFlow
        """
        self._cc._post("/1/gateways/%s/event/reboot" % (self.gateway_id))

    def ping(self, ping_id=None):
        """
        Send ping event with optional ping_id.
        EdgeFlow will respond with status message with subsystem="ping"
        """
        event = {"timestamp": time()}
        if ping_id is not None:
            event['ping_id'] = ping_id
        self._cc._post("/1/gateways/%s/event/ping" % self.gateway_id, json=event)

    def trigger_camera_capture(self, subject_uid, trigger_domain_unit=None):
        """
        trigger a camera capture app via the public-api
        subject_uid of the 'trigger subject' is required
        """
        event = {'subject_uid': subject_uid}
        if trigger_domain_unit is not None:
            event['trigger_domain_unit'] = trigger_domain_unit
        self._cc._post("/1/gateways/%s/event/trigger_camera_capture" % self.gateway_id, json=event)

    def status(self, subsystem_name=None, start=None, end=None, reverse=True, limit=None):
        """
        Yields EdgeFlow status, optionally only for a particular subsytem, sorted by timestamp.

        start (float)          filter by last update timestamp > start (seconds since epoch)
        end (float)            filter by last update timestamp < end   (seconds since epoch)
        reverse (bool)         reverse the sorting order: sort high to low
        limit (int)            yield maximum of limit results
        """
        args = []
        if start is not None:
            args.append("start=%f" % start)
        if end is not None:
            args.append("end=%f" % end)
        if reverse:
            args.append('reverse=True')
        if limit:
            assert(limit > 0)
            args.append('limit=%d' % min(limit, 100))  # api support max limit of 100

        if subsystem_name:
            url = "/1/gateways/%s/status/%s?" % (self.gateway_id, subsystem_name)
        else:
            url = "/1/gateways/%s/status?" % self.gateway_id

        url += "&".join(args)

        def get_next(url):
            resp = self._cc._get(url)
            return resp.json()

        count = 0
        while url:
            resp = get_next(url)
            for data in resp['data']:
                yield data
                count += 1
                if limit and count == limit:
                    return
            url = resp['paging'].get('next')



"""
helper functions for dealing with time ranges

Copyright (C) 2018 Cogniac Corporation
"""

from time import time
import datetime


class UnknownTimeframe(Exception):
    """
    unknown timeframe specification
    """


timeframes = {
    "minute":         "last minute",
    "hour":           "last hour",
    "day":            "last 24 hours until now",
    "week":           "last 7 days (7*24 hours) until now",
    "month":          "last 31 days until now",
    "previous-month": "previous calendar month",
    "-month":         "previous calendar month (alias)",
    "halfmonth":      "first 15 days of month for midmonth service usage tracking",
    "current-day":    "current calendar day until now",
    "today":          "current calendar day until now (alias)",
    "yesterday":      "previous calendar day (alias)",
    "current-month":  "current calendar month until now",
    "previous-day":   "previous calendar day"
    }


def start_end_times(timeframe_str):
    """
    return unix timestamps for the start time and end time corresponding to the
    user-supplied timeframe_str:

    "minute":         last minute
    "hour":           last hour
    "day":            last 24 hours until now
    "week":           last 7 days (7*24 hours) until now
    "month":          last 31 days until now
    "previous-month": previous calendar month
    "-month":         previous calendar month (alias)
    "halfmonth":      first 15 days of month for midmonth service usage tracking
    "current-day":    current calendar day until now
    "today"           current calendar day until now (alias)
    "current-month":  current calendar month until now
    "previous-day":   previous calendar day
    "yesterday":      previous calendar day (alias)

    """
    # TODO
    """
    "previous-week"
    "previous-day"

    """
    time_end = time()   # query filter end time
    dt_now = datetime.datetime.now()

    if timeframe_str == 'minute':
        time_start = time_end - 60
    elif timeframe_str == 'hour':
        time_start = time_end - 60*60
    elif timeframe_str == 'day':
        time_start = time_end - 24*60*60
    elif timeframe_str == 'week':
        time_start = time_end - 24*60*60*7
    elif timeframe_str == 'month':
        time_start = time_end - 24*60*60*31
    elif timeframe_str == '-month' or timeframe_str == "previous-month":
        last_month = dt_now.month-1
        if last_month == 0:
            last_month = 12
        start_year = dt_now.year
        if last_month > dt_now.month:
            start_year -= 1
        dt_start = datetime.datetime(start_year, last_month, 1)
        dt_end = datetime.datetime(dt_now.year, dt_now.month, 1)
        time_start = float(dt_start.strftime('%s'))
        time_end = float(dt_end.strftime('%s'))
    elif timeframe_str == 'halfmonth':
        dt_start = datetime.datetime(dt_now.year, dt_now.month, 1)
        dt_end = datetime.datetime(dt_now.year, dt_now.month, 16)
        time_start = float(dt_start.strftime('%s'))
        time_end = float(dt_end.strftime('%s'))
    elif timeframe_str == 'today' or timeframe_str == "current-day":
        dt_start = datetime.datetime(dt_now.year, dt_now.month, dt_now.day, 0, 0, 0)
        dt_end = dt_now
        time_start = float(dt_start.strftime('%s'))
        time_end = float(dt_end.strftime('%s'))
    elif timeframe_str == "current-month":
        dt_start = datetime.datetime(dt_now.year, dt_now.month, 1, 0, 0, 0)
        dt_end = dt_now
        time_start = float(dt_start.strftime('%s'))
        time_end = float(dt_end.strftime('%s'))
    elif timeframe_str == 'yesterday' or timeframe_str == "previous-day":
        dt_start = datetime.date.today() - datetime.timedelta(days=1)
        dt_end = datetime.date.today()
        time_start = float(dt_start.strftime('%s'))
        time_end = float(dt_end.strftime('%s'))
    else:
        raise UnknownTimeframe(timeframe_str)

    return time_start, time_end



def help():
    for period in timeframes:
        time_start, time_end = start_end_times(period)
        print "%20s  %s  %s  %s" % (period, datetime.datetime.fromtimestamp(time_start).strftime('%Y-%m-%d %H:%M'), datetime.datetime.fromtimestamp(time_end).strftime('%Y-%m-%d %H:%M'), timeframes[period])
    

if __name__ == "__main__":
    help()



"""
CogniacOpsReview Object Client

Copyright (C) 2016 Cogniac Corporation
"""

from retrying import retry
import six
import sys
from .common import server_error


##
#  Cogniac Ops Review
##
@six.python_2_unicode_compatible
class CogniacOpsReview(object):
    """
    CogniacOpsReview

    Record results of user judgement using cogniac view tool
    Serves as a system of record for those results as well as a point of comparison against results within the Cogniac system.
    """

    ##
    #  create
    ##
    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def create(cls,
               connection,
               review_items,
               review_unit=None):
        """
        Add an item to operations review queue:

        connnection (CogniacConnection): Authenticated CogniacConnection object
        review_items: list of
            media_id: required
            detection_ids: optional list of processing records leading to this review
            detections: optional list of detections to highlight for user
                            subject_uid
                            probability
                            app_data_type
                            app_data

        review_unit: optional str, user specified identifier associated with the review of this
        media or group of media.  Often a domain_unit would be a
        natural choice	for this id but it is not required to be a domain
        unit. The only use for this field is for subsequent searching.
        """
        data = {'review_items': review_items}
        if review_unit:
            data['review_unit'] = review_unit

        resp = connection._post("/1/ops/review", json=data)
        return CogniacOpsReview(connection, resp.json())

    ##
    #  get
    ##
    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def get(cls,
            connection):
        """
        get single item from operations review queue
        connnection (CogniacConnection): Authenticated CogniacConnection object

        return:
            review_id, cogniac supplied unique ID,
            review_items
            review_unit if any
        """
        resp = connection._get("/1/ops/review")
        return CogniacOpsReview(connection, resp.json())

    ##
    #  get pending
    ##
    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def get_pending(cls,
                    connection):
        """
        return number of pending items in operations review queue
        connnection (CogniacConnection): Authenticated CogniacConnection object
        """
        resp = connection._get("/1/ops/review/pending")
        return resp.json()['pending']

    ##
    #  update with result
    ##
    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def create_result(cls,
                      connection,
                      review_id,
                      result,
                      comment=None):
        """
        Report user judgement as result back to cogniac system for the given review_id

        connnection (CogniacConnection): Authenticated CogniacConnection object
        review_id (str): cogniac supplied unique ID, required.
        result (str): required user input, 'OK' or 'NG'
        """

        data = dict(review_id=review_id, result=result)
        if comment:
            data['comment'] = comment
        resp = connection._post("/1/ops/results", json=data)
        return CogniacOpsReview(connection, resp.json())

    ##
    #  search
    ##
    @classmethod
    def search(cls,
               connection,
               review_unit=None,
               media_id=None,
               external_media_id=None,
               result=None,
               start=None,
               end=None,
               reverse=True,
               limit=None):
        """
        search CogniacOpsReviewResults by time period
        filtered by media_id, review_unit, result

        connnection (CogniacConnection): Authenticated CogniacConnection object

        start(Float):               start time of the search period
        end(Float):                 end time of the search period
        reverse(Bool):                   order of returning results based on timestamp
        limit (int):                      max number of results to return
        """

        # build the search args
        # perform only one search at a time
        args = []
        if start is not None:
            args.append("start=%f" % start)
        if end is not None:
            args.append("end=%f" % end)

        if reverse:
            args.append('reverse=True')
        if limit:
            assert(limit > 0)
            args.append('limit=%d' % min(limit, 100))  # api support max limit of 100

        if media_id is not None:
            args.append("media_id=%s" % media_id)
        if external_media_id is not None:
            args.append("external_media_id=%s" % external_media_id)
        if review_unit is not None:
            args.append("review_unit=%s" % review_unit)
        if result is not None:
            args.append("result=%s" % result)
        url = "/1/ops/results?"
        url += "&".join(args)

        @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
        def get_next(url):
            resp = connection._get(url)
            return [CogniacOpsReview(connection, s) for s in resp.json()['data']], resp.json()['paging']

        count = 0
        while url:
            reviews, paging = get_next(url)
            for review in reviews:
                yield review
                count += 1
                if limit and count == limit:
                    return
            url = paging.get('next')

    ##
    #  __init__
    ##
    def __init__(self, connection, review_dict):
        """
        create a CogniacOpsReview

        This is not normally called directly by users, instead use:
        CogniacOpsReview.create()
        """
        self._cc = connection
        self._sub_keys = review_dict.keys()
        for k, v in review_dict.items():
            super(CogniacOpsReview, self).__setattr__(k, v)

    ##
    #  delete
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def delete(self):
        """
        Delete the op review result.
        """
        resp = self._cc._delete("/1/ops/review/%s" % self.review_id)

        for k in self._sub_keys:
            delattr(self, k)
        self._sub_keys = None
        self.connection = None

    def __setattr__(self, name, value):
        super(CogniacOpsReview, self).__setattr__(name, value)

    def __str__(self):
        return "%s" % (self.review_id)

    def __repr__(self):
        return self.__str__()



"""
CogniacUser Object

Copyright (C) 2019 Cogniac Corporation

"""

import json
from retrying import retry
from .common import *


mutable_keys = ['given_name', 'surname', 'title']


##
#   CogniacUser
##
class CogniacUser(object):

    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def get(cls, connection):
        resp = connection._get("/1/users/current")
        return CogniacUser(connection, json.loads(resp.content))

    def __init__(self, connection, tenant_dict):
        self._cc = connection
        for k, v in tenant_dict.items():
            super(CogniacUser, self).__setattr__(k, v)

    def __str__(self):
        return "%s %s (%s)" % (self.given_name, self.surname, self.email)

    def __repr__(self):
        return "%s %s (%s)" % (self.given_name, self.surname, self.email)

    def __setattr__(self, name, value):
        if name in mutable_keys:
            data = {name: value}
            resp = self._cc._post("/1/users/%s" % self.user_id, json=data)
            for k, v in resp.json().items():
                super(CogniacUser, self).__setattr__(k, v)
            return
        super(CogniacUser, self).__setattr__(name, value)

    def api_keys(self):
        resp = self._cc._get("/1/users/%s/apiKeys" % self.user_id)
        return resp.json()['data']

    def api_key(self, key_id):
        resp = self._cc._get("/1/users/%s/apiKeys/%s" % (self.user_id, key_id))
        return resp.json()

    def create_api_key(self, description):
        resp = self._cc._post("/1/users/%s/apiKeys" % self.user_id, json={'description': description})
        return resp.json()

    def delete_api_key(self, key_id):
        self._cc._delete("/1/users/%s/apiKeys/%s" % (self.user_id, key_id))



"""
CogniacExternalResult Object Client

Copyright (C) 2016 Cogniac Corporation
"""

from retrying import retry
import six
import sys
from .common import server_error


##
#  Cogniac External Result
##
@six.python_2_unicode_compatible
class CogniacExternalResult(object):
    """
    CogniacExternalResult

    Record results of any external inspections that are happening outside the system.
    Serves as a system of record for those results as well as a point of comparison against results within the Cogniac system.
    """

    ##
    #  create
    ##
    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def create(cls,
               connection,
               result_type,
               result,
               media_id=None,
               domain_unit=None):
        """
        Create a CogniacExternalResult

        connnection (CogniacConnection): Authenticated CogniacConnection object
        result_type (String):            User defined name for the results, e.g. 'operator_ok_ng'
        result (String):                 User defined json string for the result, e.g. result='ok' or result='ng'
        media_id(String):                Media_id associated with the result if inspection is done on per media basis
        domain_unit(String):             Domain_unit associated with the result if inspection is done on per part basis
        """

        data = dict(result_type=result_type, result=result)
        if media_id:
            data['media_id'] = media_id
        if domain_unit:
            data['domain_unit'] = domain_unit

        resp = connection._post("/1/externalResults", json=data)
        return CogniacExternalResult(connection, resp.json())

    ##
    #  get
    ##
    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def get(cls,
            connection,
            external_result_id):
        """
        return an existing CogniacExternalResult

        connnection (CogniacConnection): Authenticated CogniacConnection object
        external_result_id (String):     The id of the Cogniac External Result to return
        """
        resp = connection._get("/1/externalResults/%s" % external_result_id)
        return CogniacExternalResult(connection, resp.json())

    ##
    #  search
    ##
    @classmethod
    def search(cls, connection, media_id=None, domain_unit=None, time_start=None, time_end=None, reverse=True, limit=None):
        """
        search CogniacExternalResults in 3 ways: by media_id, domain_unit or time period, exactly one way should be specified.

        connnection (CogniacConnection): Authenticated CogniacConnection object

        media_id (String):               media_id to be looked up
        domain_unit(String):             domain_unit to be looked up

        time_start(Float):               start time of the search period
        time_end(Float):                 end time of the search period
        reverse(Bool):                   order of returning results based on timestamp
        limit (int):                      max number of results to return

        """

        # build the search args
        # perform only one search at a time
        data = {}
        if media_id:
            data['media_id'] = media_id
        elif domain_unit:
            data['domain_unit'] = domain_unit
        else:
            data['reverse'] = reverse
            if time_start:
                data['time_start'] = time_start
            if time_end:
                data['time_end'] = time_end
            if limit:
                data['limit'] = limit

        resp = connection._get("/1/externalResults", json=data)
        print(resp.json())
        subs = resp.json()['data']
        return [CogniacExternalResult(connection, s) for s in subs]

    ##
    #  __init__
    ##
    def __init__(self, connection, external_result_dict):
        """
        create a CogniacExternalResult

        This is not normally called directly by users, instead use:
        CogniacExternalResult.create()
        """
        self._cc = connection
        self._sub_keys = external_result_dict.keys()
        for k, v in external_result_dict.items():
            super(CogniacExternalResult, self).__setattr__(k, v)

    ##
    #  delete
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def delete(self):
        """
        Delete the external result.
        """
        resp = self._cc._delete("/1/externalResults/%s" % self.external_result_id)

        for k in self._sub_keys:
            delattr(self, k)
        self._sub_keys = None
        self.connection = None

    def __setattr__(self, name, value):
        super(CogniacExternalResult, self).__setattr__(name, value)

    def __str__(self):
        return "%s" % (self.external_result_id)

    def __repr__(self):
        return self.__str__()



from IPython.core.magic import (register_line_magic, register_cell_magic,
                                register_line_cell_magic)

import cogniac
from tabulate import tabulate
from datetime import datetime
import os
import time_range
from sys import argv

__builtins__['cc'] = None
__builtins__['S'] = None


def print_tenants(tenants):
    tenants.sort(key=lambda x: x['name'])
    data = [['tenant_id', 'name']]
    for tenant in tenants:
        data.append([tenant['tenant_id'], tenant['name']])
    print tabulate(data, headers='firstrow')


@register_line_magic
def tenants(line):
    tenants = cogniac.CogniacConnection.get_all_authorized_tenants()['tenants']
    print_tenants(tenants)

print "added ipython magic %tenants"


class Subjects(object):
    def __init__(self):
        count = 0
        for subject in cc.get_all_subjects():
            key = subject.subject_uid.replace('-', '_')  # workaround for some unclean legacy subject_uid
            self.__setattr__(key, subject)
            count += 1
        print 'added', count, 'subjects'


@register_line_magic
def authenticate(tenant_id):
    """
    authenticate to the specified tenant_id
    store CogniacConnection in cc object
    load all Cogniac Subjects into S object
    """
    cc = cogniac.CogniacConnection(tenant_id=tenant_id)
    __builtins__['cc'] = cc  # workaround ipython silliness
    print cc.tenant
    print "Adding all subjects to S"
    S = Subjects()
    __builtins__['S'] = S
    print "Type S.<tab> to autocomplete subjects"
print "added ipython magic %authenticate"


def print_detections(detections):
    # remove None values from dict
    detections = [{k: v for k, v in d.items() if v is not None} for d in detections]

    detections.sort(key=lambda x: x['created_at'])

    for d in detections:
        if 'activation' in d:
            del d['activation']
        value = datetime.fromtimestamp(d['created_at'])
        d['created_at'] = value.strftime('%Y-%m-%d %H:%M:%S')
    print tabulate(detections, headers='keys')


@register_line_magic
def media_detections(media_id):
    "print media detections for the specified media_id"
    try:
        media = cc.get_media(media_id)
    except:
        print "media_id %s not found" % media_id
        return
    print_detections(media.detections())
print "added ipython magic %media_detections"


def print_subjects(media_subjects):
    subjects = [ms['subject'] for ms in media_subjects]
    subjects = [{k: v for k, v in s.items() if v is not None} for s in subjects]
    subjects.sort(key=lambda x: x['updated_at'])
    for s in subjects:
        if 'timestamp' in s:
            del s['timestamp']
        value = datetime.fromtimestamp(s['updated_at'])
        s['updated_at'] = value.strftime('%Y-%m-%d %H:%M:%S')
    print tabulate(subjects, headers='keys')


@register_line_magic
def media_subjects(media_id):
    "print subject media associations for the specified media_id"
    try:
        media = cc.get_media(media_id)
    except:
        print "media_id %s not found" % media_id
        return

    print_subjects(media.subjects())
print "added ipython magic %media_subjects"


@register_line_magic
def users(line):
    def user_to_list(u):
        try:
            last = datetime.fromtimestamp(float(u['last_auth']))
            last = last.strftime('%Y-%m-%d %H:%M:%S')
        except:
            last = ""
        return [u['given_name'] + " " + u['surname'], u['email'], u['role'], last, u['user_id']]
    print "Users for tenant %s (%s)" %  (cc.tenant.name, cc.tenant.tenant_id)
    users = cc.tenant.users()
    users.sort(key=lambda x: x['last_auth'])
    data = [['name', 'email', 'tenant_role', 'last_auth', 'user_id']]
    for user in users:
        data.append(user_to_list(user))
    print tabulate(data, headers='firstrow')



@register_line_magic
def timeranges(line):
    """
    print list of valid timeframe selector strings, their corresponding current values, and description
    """
    time_range.help()
print "added ipython magic %timeranges"


def tenant_usage_convert_for_display(ur):
    value = datetime.fromtimestamp(ur['start_time'])
    ur['start_time'] = value.strftime('%Y-%m-%d %H:%M:%S')
    value = datetime.fromtimestamp(ur['end_time'])
    ur['end_time'] = value.strftime('%Y-%m-%d %H:%M:%S')
    ur['app_count'] = len(ur['active_model_apps'])
    gb = float(ur.get('media_bytes', 0)) / 1e9
    if gb < 1000:
        ur['media_GB'] = round(gb, 1)
    else:
        ur['media_GB'] = round(gb, 0)
    if 'media_count' not in ur:
        ur['media_count'] = 0


@register_line_magic
def usage(line):

    if line in time_range.timeframes:
        timerange_str = line
        period = None
    elif ' ' in line:
        timerange_str, period = line.split(' ')
    else:
        timerange_str, period  = "day", "hour"

    start_time, end_time = time_range.start_end_times(timerange_str)
    print 'tenant id:\t', cc.tenant.tenant_id
    print 'tenant name:\t', cc.tenant.name
    print 'report start\t', datetime.fromtimestamp(start_time).strftime('%Y-%m-%d %H:%M:%S')
    print 'report end\t', datetime.fromtimestamp(end_time).strftime('%Y-%m-%d %H:%M:%S')
    print

    if not period:
        if end_time - start_time >= (60*60*24*7):
            period = "day"
        elif end_time - start_time >= (60*60*6):
            period = 'hour'
        else:
            period = '15min'

    usage = list(cc.tenant.usage(start_time, end_time, period=period))
    for ur in usage:
        tenant_usage_convert_for_display(ur)

    tenant_headers = ['start_time', 'end_time', 'amu', 'model_outputs', 'user_feedback', 'other_outputs', 'app_count', 'media_count', 'media_GB']

    data = [tenant_headers] + [[d.get(h) for h in tenant_headers] for d in usage]
    print tabulate(data, headers='firstrow')
print "added ipython magic %usage"



@register_line_magic
def login(tname):
    """
    attempt to match user supplied partial tenant name or tenant_id 
    authenticate with the matched tenant
    """
    tenant_list = cogniac.CogniacConnection.get_all_authorized_tenants()['tenants']
    def match(t):
        return tname.lower() in t['name'].lower() or tname in t['tenant_id']
    filter_tenant_list = filter(match, tenant_list)
    if len(filter_tenant_list) == 1:
        authenticate(filter_tenant_list[0]['tenant_id'])  # use tenant from command line
    elif len(filter_tenant_list) > 1:
        print_tenants(filter_tenant_list)  # show tenants that match
    else:
        print_tenants(tenant_list)  # show all tenants


if 'COG_TENANT' in os.environ:
    tenant_id = os.environ['COG_TENANT']
    print "found COG_TENANT %s" % tenant_id
    authenticate(tenant_id)
else:
    login(argv[-1])



#!/usr/bin/env python
"""
Cogniac API Python SDK

Copyright (C) 2016 Cogniac Corporation.

"""

import logging
import os
import re
import requests
import six
import sys
from retrying import retry
from requests.auth import HTTPBasicAuth
from requests.packages.urllib3 import Retry
from requests.adapters import HTTPAdapter

from .common import server_error, raise_errors, CredentialError, credential_error

from .app     import CogniacApplication
from .subject import CogniacSubject
from .tenant  import CogniacTenant
from .user    import CogniacUser
from .media   import CogniacMedia
from .edgeflow import CogniacEdgeFlow

from .network_camera import CogniacNetworkCamera

logger = logging.getLogger(__name__)


###
#  CogniacConnection
##
class CogniacConnection(object):
    """
    CogniacConnection

    Authenticate to the Cogniac System and maintain session state.

    CogniacConnection also provides helper functions for creating and retrieving the user's
    CogniacMedia, CogniacApplication, CogniacSubject, and CogniacTenant objects.

    If a user is a member of multiple tenants the user can retrieve his list of associated
    tenants via the CogniacConnection.get_all_authorized_tenants() classmethod.
    """

    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def get_all_authorized_tenants(cls, username=None, password=None, url_prefix="https://api.cogniac.io/"):
        """
        return the list of valid tenants for the specified user credentials and url_prefix
        """
        if 'COG_API_KEY' in os.environ:
            resp = requests.get(url_prefix + "/1/users/current/tenants",
                                headers={"Authorization": "Key %s" % os.environ['COG_API_KEY']})
            raise_errors(resp)
            return resp.json()

        if username is None and password is None:
            # credentials not specified, use environment variables if found
            try:
                username = os.environ['COG_USER']
                password = os.environ['COG_PASS']
            except:
                raise Exception("No Cogniac Credentials. Try setting COG_USER and COG_PASS environment.")

        resp = requests.get(url_prefix + "/1/users/current/tenants", auth=HTTPBasicAuth(username, password))
        raise_errors(resp)
        return resp.json()

    def __init__(self,
                 username=None,
                 password=None,
                 api_key=None,
                 tenant_id=None,
                 timeout=60,
                 url_prefix="https://api.cogniac.io/"):
        """
        Create an authenticated CogniacConnection with the following credentials:

        username (String):            The Cogniac account username (usually an email address).
                                      The username can also be supplied via the COG_USER
                                      environment variable.

        password (String):            The associated Cogniac account password.
                                      The password can also be supplied via the COG_PASS
                                      environment variable.

        api_key (String):             A Cogniac-issued API key that can be used as a substitute for
                                      a username+password.  The api_key can also be supplied via the
                                      COG_API_KEY environment variable.

        tenant_id (String):           Cogniac tenant_id with which to assume credentials.
                                      This is only required if the user is a member of multiple tenants.
                                      If tenant_id is None, and the user is a member of multiple tenants
                                      then use the contents of the COG_TENANT environment variable
                                      will be used as the tenant.

        url_prefix (String):          Cogniac API url prefix.
                                      Defaults to "https://api.cogniac.io/" for the Cogniac cloud system.
                                      If you are accessing an 'on-prem' version of the Cogniac system,
                                      please set this accordingly (e.g. 'https://your_company_name.local.cogniac.io/'
                                      or a custom DNS prefix assigned by your internal IT.)
                                      The url_prefix can alternatively be set via the COG_URL_PREFIX environment variable.

        If a user is a member of multiple tenants the user can retrieve his list of associated
        tenants via the CogniacConnection.get_all_authorized_tenants() classmethod.
        """
        self.api_key = None
        if api_key is not None:
            self.api_key = api_key
        elif username is not None and password is not None:
            self.username = username
            self.password = password
        elif 'COG_API_KEY' in os.environ:
            self.api_key = os.environ['COG_API_KEY']
        else:
            # credentials not specified, use environment variables if found
            try:
                username = os.environ['COG_USER']
                password = os.environ['COG_PASS']
                self.username = username
                self.password = password
            except:
                raise Exception("No Cogniac Credentials. Specify username and password or set COG_USER, COG_PASS or COG_API_KEY environment variables.")

        if 'COG_URL_PREFIX' in os.environ:
            url_prefix = os.environ['COG_URL_PREFIX']
        m = re.search(r'/\d+(/)?$', url_prefix)
        # Strip API version number and tailing '/' from URL prefix.
        if m is not None:
            url_prefix = url_prefix[0:m.span()[0]]
        if url_prefix.endswith('/'):
            url_prefix = url_prefix[0:-1]

        self.url_prefix = url_prefix
        self.timeout = timeout

        logger.info("Connecting to Cogniac system at %s" % url_prefix)

        if tenant_id is None:
            try:
                tenant_id = os.environ['COG_TENANT']
            except:
                if self.api_key:
                    print("tenant_id must be explicitly specified when using api_key")
                    raise Exception("Unspecified tenant")

                # get list of user's tenants
                tenants = CogniacConnection.get_all_authorized_tenants(username, password, url_prefix)['tenants']
                if len(tenants) == 1:
                    # only one choice -- automatically use that
                    tenant_id = tenants[0]['tenant_id']
                else:
                    # try to be helpful and provider interactive user with a list of valid tenants
                    print("\nError: must specify tenant (e.g. export COG_TENANT=... ) from the following choices:")
                    tenants.sort(key=lambda x: x['name'])
                    for tenant in tenants:
                        print("%24s (%s)    export COG_TENANT='%s'" % (tenant['name'], tenant['tenant_id'], tenant['tenant_id']))
                    print
                    raise Exception("Unspecified tenant")

        self.tenant_id = tenant_id

        # get and store auth headers
        self.__authenticate()

        # get tenant and user objects associated with this connection
        self.tenant = CogniacTenant.get(self)
        self.user = CogniacUser.get(self)

        if self.tenant.region is not None:
            # use tenant object's specified region preference
            # print "Using API endpoint from Tenant:", self.tenant.region
            self.url_prefix = 'https://' + self.tenant.region

    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def __authenticate(self):
        #  Authenticate to the cogniac system using a username and password or an API KEY
        #  Save the http Authorization headers that can be used for subsequent http requests to the cogniac API.
        tenant_data = {"tenant_id": self.tenant_id}
        if self.api_key:
            # trade API KEY for user+tenant token
            resp = requests.get(self.url_prefix + "/1/token",
                                params=tenant_data,
                                headers={"Authorization": "Key %s" % self.api_key},
                                timeout=self.timeout)
        else:
            # trade username/password for user+tenant token

            # https://staging.cogniac.io/21/users/mfa/status
            resp = requests.get(self.url_prefix + "/21/users/mfa/status",
                                auth=HTTPBasicAuth(self.username, self.password),
                                timeout=self.timeout)
            raise_errors(resp)

            mfa_status = resp.json()
            if mfa_status.get('totp') == 'active':
                try:
                    totp = six.moves.input('Multi-Factor Authentication is required. Enter OTP: ')
                except KeyboardInterrupt:
                    sys.exit()
                tenant_data['otp'] = totp

            resp = requests.get(self.url_prefix + "/1/token",
                                params=tenant_data,
                                auth=HTTPBasicAuth(self.username, self.password),
                                timeout=self.timeout)

        raise_errors(resp)

        token = resp.json()
        headers = {"Authorization": "Bearer %s" % token['access_token']}
        self.session = requests.Session()
        # configure session with appropriate retries
        self.session.mount('https://', HTTPAdapter(max_retries=Retry(connect=5,
                                                                     read=5,
                                                                     status=5,
                                                                     redirect=2,
                                                                     backoff_factor=.001,
                                                                     status_forcelist=(500, 502, 503, 504))))
        self.session.headers.update(headers)

    @retry(stop_max_attempt_number=3, retry_on_exception=credential_error)
    def _head(self, url, timeout=None, **kwargs):
        """
        wrap requests session to re-authenticate on credential expiration
        """
        if not url.startswith("http"):
            # Prepend /1/ version if no version is specified in the URL (backward compatibility).
            m = re.search(r'^/\d+(/)?', url)
            if m is None:
                url = '/1' + url

            url = self.url_prefix + url
        if timeout is None:
            timeout = self.timeout
        try:
            resp = self.session.head(url, timeout=timeout, **kwargs)
            raise_errors(resp)
        except CredentialError:
            self.__authenticate()
            raise
        return resp

    @retry(stop_max_attempt_number=3, retry_on_exception=credential_error)
    def _get(self, url, timeout=None, **kwargs):
        """
        wrap requests session to re-authenticate on credential expiration
        """
        if not url.startswith("http"):
            # Prepend /1/ version if no version is specified in the URL (backward compatibility).
            m = re.search(r'^/\d+(/)?', url)
            if m is None:
                url = '/1' + url

            url = self.url_prefix + url
        if timeout is None:
            timeout = self.timeout
        try:
            resp = self.session.get(url, timeout=timeout, **kwargs)
            raise_errors(resp)
        except CredentialError:
            self.__authenticate()
            raise
        return resp

    @retry(stop_max_attempt_number=3, retry_on_exception=credential_error)
    def _post(self, url, timeout=None, **kwargs):
        """
        wrap requests session to re-authenticate on credential expiration
        """
        if not url.startswith("http"):
            # Prepend /1/ version if no version is specified in the URL (backward compatibility).
            m = re.search(r'^/\d+(/)?', url)
            if m is None:
                url = '/1' + url

            url = self.url_prefix + url
        if timeout is None:
            timeout = self.timeout
        try:
            resp = self.session.post(url, timeout=timeout, **kwargs)
            raise_errors(resp)
        except CredentialError:
            self.__authenticate()
            raise
        return resp

    @retry(stop_max_attempt_number=3, retry_on_exception=credential_error)
    def _delete(self, url, timeout=None, **kwargs):
        """
        wrap requests session to re-authenticate on credential expiration
        """
        if not url.startswith("http"):
            # Prepend /1/ version if no version is specified in the URL (backward compatibility).
            m = re.search(r'^/\d+(/)?', url)
            if m is None:
                url = '/1' + url

            url = self.url_prefix + url
        if timeout is None:
            timeout = self.timeout
        try:
            resp = self.session.delete(url, timeout=timeout, **kwargs)
            raise_errors(resp)
        except CredentialError:
            self.__authenticate()
            raise
        return resp

    def get_tenant(self):
        """
        return the currently authenticated CogniacTenant
        """
        return self.tenant

    def get_all_applications(self):
        """
        return CogniacApplications for all applications belonging to the currently authenticated tenant
        """
        return CogniacApplication.get_all(self)

    def get_application(self, application_id):
        """
        return an existing CogniacApplication

        application_id (String):             The application_id of the Cogniac application to return
        """
        return CogniacApplication.get(self, application_id)

    def create_application(self,
                           name,
                           application_type,
                           description=None,
                           active=True,
                           input_subjects=None,
                           output_subjects=None,
                           app_managers=None):
        """
        Create a new CogniacApplication

        name (String):                       Name of new application
        application_type (String)            Cogniac Application Type name
        description (String):                Optional description of new application
        active (Boolean):                    Application operational state
        input_subjects ([CogniacSubjects]):  List of CogniacSubjects inputs to this application
        output_subjects ([CogniacSubjects]): List of CogniacSubjects outputs for this application
        """
        return CogniacApplication.create(self,
                                         name=name,
                                         application_type=application_type,
                                         description=description,
                                         active=active,
                                         input_subjects=input_subjects,
                                         output_subjects=output_subjects,
                                         app_managers=app_managers)

    def get_all_subjects(self, public_read=False, public_write=False):
        """
        return CogniacSubjects for all subjects belonging to the currently authenticated tenant
        """
        return CogniacSubject.get_all(self, public_read=public_read, public_write=public_write)

    def search_subjects(self, ids=[], prefix=None, similar=None, name=None, tenant_owned=True, public_read=False, public_write=False, limit=10):
        """
        return CogniacSubjects based on given search filters
        """
        return CogniacSubject.search(self,
                                     ids=ids,
                                     prefix=prefix,
                                     similar=similar,
                                     name=name,
                                     tenant_owned=tenant_owned,
                                     public_read=public_read,
                                     public_write=public_write,
                                     limit=limit)

    def get_subject(self, subject_uid):
        """
        return an existing CogniacSubject

        subject_id (String):                 The subject_id of the Cogniac Subject to return
        """
        return CogniacSubject.get(self, subject_uid)

    def create_subject(self,
                       name,
                       description=None,
                       external_id=None,
                       public_read=False,
                       public_write=False):
        """
        Create a CogniacSubject

        name (String):                       Name of new subject
        description (String):                Optional description of the subject
        public_read(Bool):                   Subject media is accessible to other tenants and can be input into other tenant's apps.
        public_write(Bool):                  Other tenants can access and associate media with this subject.
        """
        return CogniacSubject.create(self,
                                     name=name,
                                     description=description,
                                     external_id=external_id,
                                     public_read=public_read,
                                     public_write=public_write)

    def get_media(self, media_id):
        """
        return a CogniacMedia object for an existing media item

        connnection (CogniacConnection):     Authenticated CogniacConnection object
        media_id (String):                   The media_id of the Cogniac Media item to return
        """
        return CogniacMedia.get(self, media_id)

    def search_media(self, md5=None, filename=None, external_media_id=None, domain_unit=None, limit=None):
        """
        return list of CogniacMedia within tenant based on specified md5, filename, or external_media_id
        """
        return CogniacMedia.search(self, md5, filename, external_media_id, domain_unit, limit)

    def create_media(self,
                     filename,
                     meta_tags=None,
                     force_set=None,
                     set_assignment=None,
                     external_media_id=None,
                     original_url=None,
                     original_landing_url=None,
                     license=None,
                     author_profile_url=None,
                     author=None,
                     title=None,
                     media_timestamp=None,
                     domain_unit=None,
                     trigger_id=None,
                     sequence_ix=None,
                     custom_data=None,
                     fp=None):
        """
        Create a new CogniacMedia object and upload the media to the Cogniac System.

        filename (str):                   Local filename or http/s URL of image or video media file
        meta_tags ([str]):                Optional list of arbitrary strings to associate with the media
        force_set (str):                  [DEPRECATED] Optionally force the media into the 'training', 'validation' or 'test' sets
        set_assignment (str):             Optionally associate media with the 'training', 'validation' or 'test' sets
        external_media_id (str):          Optional arbitrary external id for this media
        original_url(str):                Optional source url for this media
        original_landing_url (str):       Optional source landing url for this media
        license (str):                    Optional copyright licensing info for this media
        author_profile_url (str):         Optional media author url
        author (str):                     Optional author name
        title (str):                      Optional media title
        media_timestamp (float):          Optional actual timestamp of media creation/occurance time
        domain_unit (str):                Optional domain id (e.g. serial number) for set assignment grouping.
                                          Media with the same domain_unit will always be assigned to the same
                                          training or validation set. Set this to avoid overfitting when you have
                                          multiple images of the same thing or almost the same thing.
        trigger_id (str):                 Unique trigger identifier leading to a media sequence containing this media
        sequence_ix (str):                The index of this media within a triggered sequence
        custom_data (str):                Opaque user-specified data associated with this media; limited to 32KB
        fp (file):                        A '.read()'-supporting file-like object (under 16MB) from which to acquire
                                          the media instead of reading the media from the specified filename.
        """
        return CogniacMedia.create(self,
                                   filename=filename,
                                   meta_tags=meta_tags,
                                   force_set=force_set,
                                   set_assignment=set_assignment,
                                   external_media_id=external_media_id,
                                   original_url=original_url,
                                   original_landing_url=original_landing_url,
                                   license=license,
                                   author_profile_url=author_profile_url,
                                   author=author,
                                   title=title,
                                   media_timestamp=media_timestamp,
                                   domain_unit=domain_unit,
                                   trigger_id=trigger_id,
                                   sequence_ix=sequence_ix,
                                   custom_data=custom_data,
                                   fp=fp)

    def get_version(self, auth=False):
        """
        get api version info
        auth (bool):  use authenticated endpoint for benchmark purposes
        returns json version info
        """
        if auth:
            url = self.url_prefix + "/1/authversion"
        else:
            url = self.url_prefix + "/1/version"
        resp = self._get(url)
        raise_errors(resp)
        return resp.json()

    def get_all_cameras(self):
        """
        return CogniacNetworkCameras for all netcams belonging to the currently authenticated tenant
        """
        return CogniacNetworkCamera.get_all(self)

    def get_camera(self, network_camera_id):
        """
        return an existing CogniacNetworkCamera

        network_camera_id (String):  The id of the Cogniac network camera to return
        """
        return CogniacNetworkCamera.get(self, network_camera_id)

    def get_all_edgeflows(self):
        """
        return CogniacEdgeFlow for all EdgeFlows belonging to the currently authenticated tenant
        """
        return CogniacEdgeFlow.get_all(self)

    def get_edgeflow(self, edgeflow_id):
        """
        return an existing CogniacEdgeFlow

        edgeflow_id (String):  The id of the Cogniac EdgeFlow to return
        """
        return CogniacEdgeFlow.get(self, edgeflow_id)


if __name__ == "__main__":
    c = CogniacConnection()
    tenant = c.tenant()
    from pprint import pprint
    print(tenant)
    print("\nApplications:")
    for app in tenant.applications():
        pprint(app.name)
    print("\Subjects:")
    for sub in tenant.subjects():
        print(sub.name, sub.subject_uid)



"""
CogniacSubject Object Client

Copyright (C) 2016 Cogniac Corporation
"""

from retrying import retry
from .common import *
import six
import sys

from .media import CogniacMedia


##
#  CogniacSubject
##
@six.python_2_unicode_compatible
class CogniacSubject(object):
    """
    CogniacSubject
    Subjects are a central organizational mechanism in the Cogniac system.
    A subject is any user-defined concept that is relevant to images or video.
    More generally a subject can represent any logical grouping of images of video.

    Most Cogniac applications work by taking input media from user-defined subjects
    and outputing those media to other user-defined subjects based on the content
    of the media.

    Create a new subject with
    CogniacConnection.create_subject() or CogniacSubject.create()

    Get an existing subject with
    CogniacConnection.get_subject() or CogniacSubject.get()

    Get all tenant's subject with
    CogniacConnection.get_all_subjects() or CogniacSubject.get_all()

    Writes to mutable CogniacSubjects attributes are saved immediately via the Cogniac API.
    """

    ##
    #  create
    ##
    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def create(cls,
               connection,
               name,
               description=None,
               external_id=None,
               public_read=False,
               public_write=False):
        """
        Create a CogniacSubject

        connnection (CogniacConnection):     Authenticated CogniacConnection object
        name (String):                       Name of new subject
        description (String):                Optional description of the subject
        external_id (String):                Optional user supplied id for the subject
        public_read(Bool):                   Subject media is accessible to other tenants and can be input into other tenant's apps.
        public_write(Bool):                  Other tenants can access and associate media with this subject.
        """
        if public_write:
            public_read = True

        data = dict(name=name, public_read=public_read, public_write=public_write)
        if description:
            data['description'] = description
        if external_id:
            data['external_id'] = external_id

        resp = connection._post("/1/subjects", json=data)

        return CogniacSubject(connection, resp.json())

    ##
    #  get
    ##
    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def get(cls,
            connection,
            subject_uid):
        """
        return an existing CogniacSubject

        connnection (CogniacConnection):     Authenticated CogniacConnection object
        subject_id (String):                 The subject_id of the Cogniac Subject to return
        """
        resp = connection._get("/1/subjects/%s" % subject_uid)
        return CogniacSubject(connection, resp.json())

    ##
    #  get_all
    ##
    @classmethod
    def get_all(cls, connection, public_read=False, public_write=False):
        """
        return CogniacSubjects for all subjects belonging to the currently authenticated tenant

        connnection (CogniacConnection):     Authenticated CogniacConnection object
        public_read (bool):                  return subjects that are publicly readable (other tenants can use associated media)
        public_write (bool):                 return subjects that are publicly writeable (others tenants associate new media)
        """

        args = []
        if public_read:
            args.append("public_read=True")
        if public_write:
            args.append("public_read_write=True")

        url = "/1/tenants/%s/subjects?" % connection.tenant.tenant_id
        url += "&".join(args)

        resp = connection._get(url)
        subs = resp.json()['data']
        return [CogniacSubject(connection, s) for s in subs]

    ##
    #  search
    ##
    @classmethod
    def search(cls, connection, ids=[], prefix=None, similar=None, name=None, tenant_owned=True, public_read=False, public_write=False, limit=10):
        """
        search CogniacSubjects for either batch subject ID's, based on prefix, or semantic simlarity

        connnection (CogniacConnection):     Authenticated CogniacConnection object

        ids (list of strings):               return list of subjects with the given subject_uid's
        prefix (string):                     return subjects that have a name containing the given string;
                                             direct prefix matching results in a higher search score
        similar (string):                    return subjects with a name that is related to the given search term;
                                             i.e. similar='cat' will return a set containing the subject 'dog'
        name (string):                       returns list of subjects with the exact name specified.

        Exactly one ids, prefix, similar, or name must be specified.

        tenant_owned (bool):                 return subjects belonging to this tenant.
        public_read (bool):                  return subjects that are publicly readable (other tenants can use associated media)
        public_write (bool):                 return subjects that are publicly writeable (others tenants associate new media)
        limit (int):                         max number of results to return
        """

        args = []

        # build the search args
        args.append('limit=%d' % limit)
        args.append('tenant_read_write=%s' % str(tenant_owned))
        if public_read:
            args.append("public_read=True")

        if public_write:
            args.append("public_read_write=True")

        # perform only one search at a time
        # id search, prefix search, similarity or name search
        if len(ids):
            args.append('ids=%s' % (',').join(ids))
        elif prefix:
            args.append('prefix=%s' % prefix)
        elif similar:
            args.append('similar=%s' % similar)
        elif name:
            args.append('name=%s' % name)

        url = "/1/tenants/%s/subjects?" % connection.tenant.tenant_id
        url += "&".join(args)

        resp = connection._get(url)
        subs = resp.json()['data']
        return [CogniacSubject(connection, s) for s in subs]

    ##
    #  __init__
    ##
    def __init__(self, connection, subject_dict):
        """
        create a CogniacSubject

        This is not normally called directly by users, instead use:
        CogniacConnection.create_subject() or
        CogniacSubject.create()
        """
        self._cc = connection
        self._sub_keys = subject_dict.keys()
        for k, v in subject_dict.items():
            super(CogniacSubject, self).__setattr__(k, v)

    ##
    #  delete
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def delete(self):
        """
        Delete the subject.
        """
        resp = self._cc._delete("/1/subjects/%s" % self.subject_uid)

        for k in self._sub_keys:
            delattr(self, k)
        self._sub_keys = None
        self.connection = None

    def __setattr__(self, name, value):
        if name in ['subject_uid', 'created_at', 'created_by', 'modified_at', 'modified_by']:
            raise AttributeError("%s is immutable" % name)
        if name in ['name', 'description', 'expires_in', 'external_id', 'custom_data']:
            data = {name: value}
            resp = self._cc._post("/1/subjects/%s" % self.subject_uid, json=data)
            for k, v in resp.json().items():
                super(CogniacSubject, self).__setattr__(k, v)
            return
        super(CogniacSubject, self).__setattr__(name, value)

    def __str__(self):
        return "%s (%s)" % (self.name, self.subject_uid)

    def __repr__(self):
        return self.__str__()

    ##
    #  create_reference_media
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def create_reference_media(self,
                               filename,
                               meta_tags=None,
                               force_set=None,
                               set_assignment=None,
                               external_media_id=None,
                               original_url=None,
                               original_landing_url=None,
                               license=None,
                               author_profile_url=None,
                               author=None,
                               title=None,
                               media_timestamp=None,
                               domain_unit=None,
                               trigger_id=None,
                               sequence_ix=None,
                               custom_data=None,
                               fp=None):
        """
        Create a new CogniacMedia object, upload the media to the Cogniac
        System, and set it as the subject's reference media.

        connnection (CogniacConnection):  Authenticated CogniacConnection object
        filename (str):                   Local filename or http/s URL of image or video media file
        meta_tags ([str]):                Optional list of arbitrary strings to associate with the media
        force_set (str):                  [DEPRECATED] Optionally force the media into the 'training', 'validation' or 'test' sets
        set_assignment (str):             Optionally force the media into the 'training', 'validation' or 'test' sets
        external_media_id (str):          Optional arbitrary external id for this media
        original_url(str):                Optional source url for this media
        original_landing_url (str):       Optional source landing url for this media
        license (str):                    Optional copyright licensing info for this media
        author_profile_url (str):         Optional media author url
        author (str):                     Optional author name
        title (str):                      Optional media title
        media_timestamp (float):          Optional actual timestamp of media creation/occurance time
        domain_unit (str):                Optional domain id (e.g. serial number) for set assignment grouping.
                                          Media with the same domain_unit will always be assigned to the same
                                          training or validation set. Set this to avoid overfitting when you have
                                          multiple images of the same thing or almost the same thing.
        trigger_id (str):                 Unique trigger identifier leading to a media sequence containing this media
        sequence_ix (str):                The index of this media within a triggered sequence
        custom_data (str):                Opaque user-specified data associated with this media; limited to 32KB
        fp (file):                        A '.read()'-supporting file-like object (under 16MB) from which to acquire
                                          the media instead of reading the media from the specified filename.
        """
        args = dict()
        if meta_tags is not None:
            args['meta_tags'] = meta_tags
        if force_set is not None:
            args['force_set'] = force_set
        if set_assignment is not None:
            args['set_assignment'] = set_assignment
        if external_media_id is not None:
            args['external_media_id'] = external_media_id
        if original_url is not None:
            args['original_url'] = original_url
        if original_landing_url is not None:
            args['original_landing_url'] = original_landing_url
        if license is not None:
            args['license'] = license
        if author_profile_url is not None:
            args['author_profile_url'] = author_profile_url
        if author is not None:
            args['author'] = author
        if title is not None:
            args['title'] = title
        if media_timestamp is not None:
            args['media_timestamp'] = media_timestamp
        if domain_unit is not None:
            args['domain_unit'] = domain_unit
        if trigger_id is not None:
            args['trigger_id'] = trigger_id
        if sequence_ix is not None:
            args['sequence_ix'] = sequence_ix
        if custom_data is not None:
            args['custom_data'] = custom_data

        if filename.startswith('http'):
            args['source_url'] = filename
        elif fp is None:  # local filename
            fstat = stat(filename)
            if 'media_timestamp' not in args:
                # set the unspecified media timestamp to the earliest file time we have
                args['media_timestamp'] = file_creation_time(filename)
            if stat(filename).st_size > 12 * 1024 * 1024:
                # use the multipart interface for large files
                return CogniacMedia._create_multipart(self.connection, filename, args)

        @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
        def upload():
            if filename.startswith('http'):
                files = None
            elif fp is not None:
                fp.seek(0)  # for the retry win
                files = {'file': fp}
            else:
                files = {'file': open(filename, 'rb')}
                # api.add_resource(SubjectReferenceMedia, '/1/subjects/<string:subject_uid>/referenceMedia')
            resp = self._cc._post("/1/subjects/{}/referenceMedia".format(self.subject_uid), data=args, files=files)
            return resp

        resp = upload()
        return CogniacMedia(self._cc, resp.json())

    ##
    #  dissassociate_media
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def disassociate_media(self,
                           media,
                           focus=None):
        """
        Disassociate the media (with an optional focus within the media) with this subject.

        media (String or CogniacMedia)   media object or media_id to associate
        focus (dict)                     Optional Focus of this association within the media
                                         'frame'   Optional frame within a video media
                                         'box'     Optional dictionary of bounding box pixel offsets (with keys x0, x1, y0, y1) within the frame.

        """
        if type(media) is CogniacMedia:
            data = {'media_id': media.media_id}
        else:
            data = {'media_id': media}

        if focus is not None:
            data['focus'] = focus
        self._cc._delete("/1/subjects/%s/media" % self.subject_uid, json=data)

    ##
    #  associate_media
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def associate_media(self,
                        media,
                        focus=None,
                        consensus='None',
                        probability=None,
                        force_feedback=False,
                        force_random_feedback=False,
                        app_data=None,
                        app_data_type=None,
                        enable_wait_result=False):
        """
        Associate the media (with an optional focus within the media) with this subject.

        media (String or CogniacMedia)   media object or media_id to associate
        focus (dict)                     Optional Focus of this association within the media
                                         'frame'   Optional frame within a video media
                                         'box'     Optional dictionary of bounding box pixel offsets (with keys x0, x1, y0, y1) within the frame.
        consensus (str)                  'True' if media is associated with subject,
                                         'False' if media is NOT associated with subject.
                                         'Sidelined' if consensus was decisively not provided.
                                         'None' if consenus is unknown.  Required uncal_prob
        probability (float)              association probability, only valid if consensus = 'None'
        force_feedback (bool)            True to force feedback on the media item in downstream apps
        force_random_feedback(bool)      True to force feedback on the media item in downstream apps because.
                                         Set only if this media was randomly selected for feedback.
                                         This media will be used for the (random) performance assessment.
        app_data                         Specific app data for this subject-media association
        app_data_type                    Type of app data for the subject-media association
        enable_wait_result(bool)         When True: enable 'synchronous' result interfaces whereby subsequent
                                         GET /media/<media_id>/detections?wait_capture_id=<capture_id> endpoints block until
                                         the full results of the application pipeline resulting from this input association are available

        returns the unique capture_id
        """

        if type(media) is CogniacMedia:
            data = {'media_id': media.media_id}
        else:
            data = {'media_id': media}

        if focus is not None:
            data['focus'] = focus

        assert(consensus in ['True', 'False', 'None', 'Sidelined'])
        data['consensus'] = consensus

        if consensus is 'None' and probability is None:
            data['uncal_prob'] = .99

        if probability is not None:
            assert(consensus == 'None')
            data['uncal_prob'] = probability

        if enable_wait_result:
            data['enable_wait_result'] = True

        data['force_feedback'] = force_feedback
        data['force_random_feedback'] = force_random_feedback
        data['app_data_type'] = app_data_type
        data['app_data'] = app_data

        resp = self._cc._post("/1/subjects/%s/media" % self.subject_uid, json=data)
        return resp.json()['capture_id']

    ##
    #  media_associations
    ##
    def media_associations(self, start=None, end=None, reverse=True, probability_lower=None, probability_upper=None, consensus=None, sort_probability=False, limit=None, abridged_media=True):
        """
        yield media associations for the subject sorted by last update timestamp or probability

        start (float)          filter by last update timestamp > start (seconds since epoch)
        end (float)            filter by last update timestamp < end   (seconds since epoch)
        reverse (bool)         reverse the sorting order: sort high to low
        probability_lower:     filter by probability > probability_lower
        probability_upper:     filter by probability < probability_upper
        consensus (string):    filter by consensus label: "True", "False" or "None"
        sort_probability(bool) Sort by probability instead of last update timestamp
        limit (int)            yield maximum of limit results
        abridged_media (bool)  return full media items if False (slower), otherwise return just media_id's for each media_item

        Returns (yield) association dictionary with the following fields:

        media_id        the media_id
        subject_uid		the subject_uid
        focus (dict)    Optional Focus of this association within the media
                           'frame'   Optional frame within a video media
                           'box'     Optional dictionary of bounding box pixel offsets (with keys x0, x1, y0, y1) within the frame.
        probability		current assessment of the probability in [0,1] that the subject_uid is associated with the media_id
                        1 = subject definitely associated with media+focus
                        0 = subject definitely NOT associated with media+focus
                        0.5 ~= uncertain association between subject and media+focus
        timestamp	time of last update to subject media association
        app_data_type	optional app data type if applicable
        app_data        optional app data if applicable
        consensus		'True', 'False', or 'Sidelined', or None
                        'True' if there is consensus that the subject is associated with the media
                            (Media will be used as a positive training example of the subject)
                        'False' if there is consensus that the subject is not associated w/the media
                            (Media will be used as a negative training example of the subject.)
                        'Sidelined' if the media's association with the subject was decisively not provided. 
                         None if if there is not enough evidence to reach consensus
                         Some application types only support 'True' or None.
        """

        args = []
        if start is not None:
            args.append("start=%f" % start)
        if end is not None:
            args.append("end=%f" % end)
        if probability_lower is not None:
            args.append("probability_lower=%f" % probability_lower)
        if probability_upper is not None:
            args.append("probability_upper=%f" % probability_upper)
        if consensus is not None:
            assert(consensus in ['True', 'False', 'Sidelined', 'None'])
            args.append("consensus=%s" % consensus)
        if reverse:
            args.append('reverse=True')
        else:
            args.append('reverse=False')
        if sort_probability:
            args.append("sort=probability")
        if limit:
            assert(limit > 0)
            args.append('limit=%d' % min(limit, 100))  # api support max limit of 100
        if abridged_media:
            args.append('abridged_media=True')

        url = "/1/subjects/%s/media?" % self.subject_uid
        url += "&".join(args)

        @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
        def get_next(url):
            resp = self._cc._get(url)
            return resp.json()

        count = 0
        while url:
            resp = get_next(url)
            for sma in resp['data']:
                yield sma
                count += 1
                if limit and count == limit:
                    return
            url = resp['paging'].get('next')



"""
Cogniac API common definitions

Copyright (C) 2016 Cogniac Corporation
"""

from requests.exceptions import ConnectionError


class CredentialError(Exception):
    """Invalid Username/Password Credentials"""
    status_code = 401


class ServerError(Exception):
    """Unknown server-side error. Operation can be retried."""


class ClientError(Exception):
    """Error with the client-supplied parameters.
    Operation should not be retried with the same parameters."""
    status_code = 400

    def __init__(self, message, status_code=None):
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code

    def __str__(self):
        return self.message


def credential_error(exception):
    return isinstance(exception, CredentialError)


def server_error(exception):
    """Return True if we should retry (in this case when it's an ServerError, False otherwise"""
    return isinstance(exception, ServerError) or isinstance(exception, ConnectionError)


def raise_errors(response):
    """
    raise ServerError or ClientError based on requests response as appropriate
    """
    if response.status_code >= 500:
        msg = "ServerError (%d): %s" % (response.status_code, response.content)
        raise ServerError(msg)

    if response.status_code == 401:
        msg = "Invalid username password credentials (%d): %s" % (response.status_code, response.content)
        raise CredentialError(msg)

    if response.status_code >= 400:
        msg = "ClientError (%d): %s" % (response.status_code, response.content)
        raise ClientError(msg, response.status_code)



"""
Cogniac Network Camera Object Client

Copyright (C) 2019 Cogniac Corporation
"""

from retrying import retry
import six
import sys
from .common import server_error

camera_model_keys = ['last_pose_change_timestamp',
                     'resolution_h_px', 'resolution_v_px',
                     'pixel_h_um', 'pixel_v_um',
                     'focal_length_h_mm', 'focal_length_v_mm',
                     'skew', 'ch_px', 'cv_px',
                     'radial_distortion_coefficients',
                     'tangential_distortion_coefficients',
                     'pitch', 'yaw', 'roll',
                     'tx_m', 'ty_m', 'tz_m',
                     'origin_x', 'origin_y',
                     'x_axis_x', 'x_axis_y',
                     'y_axis_x', 'y_axis_y',
                     'z_axis_x', 'z_axis_y']

mutable_keys = ['url', 'current_IP', 'camera_name', 'description',
                'active', 'lat', 'lon', 'hae', 'alt_subject_uid', 'custom_configuration'] + camera_model_keys

immutable_keys = ['network_camera_id', 'created_at',
                  'created_by', 'modified_at', 'modified_by']


##
#  Cogniac Network Camera
##
@six.python_2_unicode_compatible
class CogniacNetworkCamera(object):
    """
    CogniacNetworkCamera
    """

    ##
    #  create
    ##
    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def create(cls,
               connection,
               name,
               url,
               description=None,
               active=True,
               discovered_by=None,
               spec_version_major=None,
               spec_version_minor=None,
               device_mode=None,
               IP_config_options=None,
               IP_config_current=None,
               current_IP=None,
               current_subnet_mask=None,
               default_gateway=None,
               mac_address=None,
               model_name=None,
               serial_number=None,
               device_version=None,
               manufacturer_name=None,
               manufacturer_info=None,
               user_defined_name=None):

        """
        Create a network camera object

        connnection (CogniacConnection): Authenticated CogniacConnection object
        name (String):                   Name of new camera
        url(String):                     url of the network camera
        description (String):            Optional description of the camera
        active(Boolean):                 True to indicate system should process the images
        discovered_by (String):          cogniac edgeflow_id of the Edgeflow that discovered this camera

        plus device dictionaries returned from GVCP discovery
        """
        data = dict(camera_name=name)
        data['url'] = url
        if active:
            data['active'] = 1
        else:
            data['active'] = 0

        if description:
            data['description'] = description

        if discovered_by:
            data['discovered_by'] = discovered_by

        if spec_version_major:
            data['spec_version_major'] = str(spec_version_major)
        if spec_version_minor:
            data['spec_version_minor'] = str(spec_version_minor)

        if device_mode:
            data['device_mode'] = str(device_mode)

        if IP_config_options:
            data['IP_config_options'] = str(IP_config_options)
        if IP_config_current:
            data['IP_config_current'] = str(IP_config_current)
        if current_IP:
            data['current_IP'] = current_IP
        if current_subnet_mask:
            data['current_subnet_mask'] = current_subnet_mask
        if default_gateway:
            data['default_gateway'] = default_gateway
        if mac_address:
            data['mac_address'] = mac_address

        if model_name:
            data['model_name'] = model_name

        if serial_number:
            data['serial_number'] = str(serial_number)

        if device_version:
            data['device_version'] = device_version

        if manufacturer_name:
            data['manufacturer_name'] = manufacturer_name

        if manufacturer_info:
            data['manufacturer_info'] = manufacturer_info

        if user_defined_name:
            data['user_defined_name'] = user_defined_name

        resp = connection._post("/1/networkCameras", json=data)
        return CogniacNetworkCamera(connection, resp.json())

    ##
    #  get
    ##
    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def get(cls,
            connection,
            network_camera_id):
        """
        get single network camera
        connnection (CogniacConnection): Authenticated CogniacConnection object
        network_camera_id (String): the netcam id of the Cogniac NetCam object
        returns CogniacNetworkCamera object
        """
        resp = connection._get("/1/networkCameras/%s" % network_camera_id)
        return CogniacNetworkCamera(connection, resp.json())

    ##
    #  get_all
    ##
    @classmethod
    def get_all(cls, connection):
        """
        return all CogniacNetworkcamera belonging to the currently authenticated tenant

        connnection (CogniacConnection):     Authenticated CogniacConnection object
        """
        resp = connection._get('/1/tenants/%s/networkCameras' % connection.tenant.tenant_id)
        netcams = resp.json()['data']
        return [CogniacNetworkCamera(connection, netcam) for netcam in netcams]

    ##
    #  post
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def update(self,
               url=None,
               current_IP=None,
               camera_name=None,
               description=None,
               active=None,
               lat=None,
               lon=None,
               hae=None,
               last_pose_change_timestamp=None,
               resolution_h_px=None,
               resolution_v_px=None,
               pixel_h_um=None,
               pixel_v_um=None,
               focal_length_h_mm=None,
               focal_length_v_mm=None,
               skew=None,
               ch_px=None,
               cv_px=None,
               radial_distortion_coefficients=None,
               tangential_distortion_coefficients=None,
               pitch=None,
               yaw=None,
               roll=None,
               tx_m=None,
               ty_m=None,
               tz_m=None,
               origin_x=None,
               origin_y=None,
               x_axis_x=None,
               x_axis_y=None,
               y_axis_x=None,
               y_axis_y=None,
               z_axis_x=None,
               z_axis_y=None):
        """
        (Deprecated)
        update single network camera's camera model
        connnection (CogniacConnection): Authenticated CogniacConnection object
        network_camera_id (String): the netcam id of the Cogniac NetCam object
        returns CogniacNetworkCamera object
        """
        print("'update' function of CogniacNetworkCamera object is deprecated "
              "and will be removed in the next version.\n"
              "Please use __setattr__ instead.\n "
              "For example:\n    netcam.current_IP = '192.164.10.1'")

        data = {}

        if url is not None:
            data['url'] = url
        if current_IP is not None:
            data['current_IP'] = current_IP
        if camera_name is not None:
            data['camera_name'] = camera_name
        if description is not None:
            data['description'] = description
        if active is not None:
            data['active'] = active
        if lat is not None:
            data['lat'] = lat
        if lon is not None:
            data['lon'] = lon
        if hae is not None:
            data['hae'] = hae

        # pose/model related
        if last_pose_change_timestamp is not None:
            data['last_pose_change_timestamp'] = last_pose_change_timestamp

        if resolution_h_px is not None:
            data['resolution_h_px'] = resolution_h_px
        if resolution_v_px is not None:
            data['resolution_v_px'] = resolution_v_px
        if pixel_h_um is not None:
            data['pixel_h_um'] = pixel_h_um
        if pixel_v_um is not None:
            data['pixel_v_um'] = pixel_v_um
        if focal_length_h_mm is not None:
            data['focal_length_h_mm'] = focal_length_h_mm
        if focal_length_v_mm is not None:
            data['focal_length_v_mm'] = focal_length_v_mm

        if skew is not None:
            data['skew'] = skew
        if ch_px is not None:
            data['ch_px'] = ch_px
        if cv_px is not None:
            data['cv_px'] = cv_px

        if radial_distortion_coefficients is not None:
            data['radial_distortion_coefficients'] = radial_distortion_coefficients
        if tangential_distortion_coefficients is not None:
            data['tangential_distortion_coefficients'] = tangential_distortion_coefficients
        if pitch is not None:
            data['pitch'] = pitch
        if yaw is not None:
            data['yaw'] = yaw
        if roll is not None:
            data['roll'] = roll

        if tx_m is not None:
            data['tx_m'] = tx_m
        if ty_m is not None:
            data['ty_m'] = ty_m
        if tz_m is not None:
            data['tz_m'] = tz_m

        if origin_x is not None:
            data['origin_x'] = origin_x
        if origin_y is not None:
            data['origin_y'] = origin_y

        if x_axis_x is not None:
            data['x_axis_x'] = x_axis_x
        if x_axis_y is not None:
            data['x_axis_y'] = x_axis_y
        if y_axis_x is not None:
            data['y_axis_x'] = y_axis_x
        if y_axis_y is not None:
            data['y_axis_y'] = y_axis_y
        if z_axis_x is not None:
            data['z_axis_x'] = z_axis_x
        if z_axis_y is not None:
            data['z_axis_y'] = z_axis_y

        resp = self._cc._post("/1/networkCameras/%s" % self.network_camera_id, json=data)
        return CogniacNetworkCamera(self._cc, resp.json())

    ##
    #  delete
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def delete(self):
        """
        Delete the op review result.
        """
        self._cc._delete("/1/networkCameras/%s" % self.network_camera_id)

        for k in self._cam_keys:
            delattr(self, k)
        self._sub_keys = None
        self.connection = None

    ##
    #  __init__
    ##
    def __init__(self, connection, netcam_dict):
        """
        create a CogniacNetworkCamera

        This is not normally called directly by users, instead use:
        CogniacNetworkCamera.create()
        """
        self._cc = connection
        self._cam_keys = netcam_dict.keys()

        for k, v in netcam_dict.items():
            super(CogniacNetworkCamera, self).__setattr__(k, v)

    def __setattr__(self, name, value):
        if name in immutable_keys:
            raise AttributeError("%s is immutable" % name)

        if name in mutable_keys:
            data = {name: value}
            print("\n====", data)
            resp = self._cc._post("/1/networkCameras/%s" % self.network_camera_id, json=data)

            for k, v in resp.json().items():
                super(CogniacNetworkCamera, self).__setattr__(k, v)

            return

        super(CogniacNetworkCamera, self).__setattr__(name, value)

    def __str__(self):
        return "%s (%s)" % (self.camera_name, self.network_camera_id)

    def __repr__(self):
        return self.__str__()



#!/usr/bin/env python

from distutils.core import setup

setup(name='cogniac',
      version='2.0.9',
      description='Python SDK for Cogniac Public API',
      packages=['cogniac'],
      author = 'Cogniac Corporation',
      author_email = 'support@cogniac.co',
      url = 'https://github.com/Cogniac/cogniac-sdk-py',
      scripts=['bin/icogniac', 'bin/cogupload'],
      install_requires=['requests', 'retrying', 'tabulate', 'six'])



"""
CogniacTenant Object

Copyright (C) 2016 Cogniac Corporation

"""

import json
import six
from retrying import retry
from .common import *


TENANT_ADMIN_ROLE = "tenant_admin"
TENANT_USER_ROLE = "tenant_user"
TENANT_VIEWER_ROLE = "tenant_viewer"
TENANT_BILLING_ROLE = "tenant_billing"


##
#   CogniacTenant
##
@six.python_2_unicode_compatible
class CogniacTenant(object):

    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def get(cls, connection):
        resp = connection._get("/1/tenants/current")
        return CogniacTenant(connection, json.loads(resp.content))

    def __init__(self, connection, tenant_dict):
        super(CogniacTenant, self).__setattr__('_tenant_keys', tenant_dict.keys())
        self._cc = connection
        for k, v in tenant_dict.items():
            super(CogniacTenant, self).__setattr__(k, v)

    def __str__(self):
        return "%s (%s)" % (self.name, self.tenant_id)

    def __repr__(self):
        return self.__str__()

    def __setattr__(self, name, value):
        if name not in self._tenant_keys:
            super(CogniacTenant, self).__setattr__(name, value)
            return
        data = {name: value}
        resp = self._cc._post("/1/tenants/%s" % self.tenant_id, json=data)
        for k, v in resp.json().items():
            super(CogniacTenant, self).__setattr__(k, v)

    def users(self):
        resp = self._cc._get("/1/tenants/%s/users" % self.tenant_id)
        return resp.json()['data']

    def set_user_role(self, user_email, role):
        users = self.users()
        users = [u for u in self.users() if u['email'] == user_email]
        if not users:
            raise Exception("unknown user_email %s" % user_email)
        data = {'user_id': users[0]['user_id'], 'role': role}
        self._cc._post("/1/tenants/%s/users/role" % self.tenant_id, json=data)

    def add_user(self, user_email, role='tenant_user'):
        users = self.users()
        users = [u for u in self.users() if u['email'] == user_email]
        if not users:
            raise Exception("unknown user_email %s" % user_email)
        data = {'user_id': users[0]['user_id'], 'role': role}
        self._cc._post("/1/tenants/%s/users" % self.tenant_id, json=data)

    def delete_user(self, user_email):
        users = self.users()
        users = [u for u in self.users() if u['email'] == user_email]
        if not users:
            raise Exception("unknown user_email %s" % user_email)
        data = {'user_id': users[0]['user_id']}
        self._cc._delete("/1/tenants/%s/users" % self.tenant_id, json=data)

    def usage(self, start, end, period='15min'):

        assert(period in ['15min', 'hour', 'day'])

        url = "/1/usage/summary?period=%s&start=%d&end=%d" % (period, start, end)

        @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
        def get_next(url):
            resp = self._cc._get(url)
            return resp.json()

        while url:
            resp = get_next(url)
            for record in resp['data']:
                yield record
            url = resp['paging'].get('next')



"""
CogniacApplication Object Client

Copyright (C) 2016 Cogniac Corporation
"""

from retrying import retry
import six
from .common import server_error


@six.python_2_unicode_compatible
class CogniacApplication(object):
    """
    CogniacApplication
    Applications are the main locus of activity within the Cogniac System.

    This classes manages applications within the Cogniac System via the
    Cogniac public API application endpoints.

    Create a new application with
    CogniacConnection.create_application() or CogniacApplication.create()

    Get an existing application with
    CogniacConnection.get_application() or CogniacApplication.get()

    Get all tenant's applications with
    CogniacConnection.get_all_applications() or CogniacApplication.get_all()

    Writes to mutable CogniacApplication attributes are saved immediately via the Cogniac API.
    """

    ##
    #  create
    ##
    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def create(cls,
               connection,
               name,
               application_type,
               description=None,
               active=True,
               input_subjects=None,
               output_subjects=None,
               app_managers=None):
        """
        Create a new CogniacApplication

        connnection (CogniacConnection):     Authenticated CogniacConnection object
        name (String):                       Name of new application
        application_type (String)            Cogniac Application Type name
        description (String):                Optional description of the application
        active (Boolean):                    Application operational state
        input_subjects ([CogniacSubjects]):  List of CogniacSubjects inputs to this application
        output_subjects ([CogniacSubjects]): List of CogniacSubjects outputs for this application
        app_managers ([String]):             List of email addresses authorized to be Application Managers
        """
        data = dict(name=name, active=active, type=application_type)

        if description:
            data['description'] = description
        if input_subjects:
            if type(input_subjects[0]) == str:
                data['input_subjects'] = input_subjects
            else:
                data['input_subjects'] = [s.subject_uid for s in input_subjects]
        if output_subjects:
            if type(output_subjects[0]) == str:
                data['output_subjects'] = output_subjects
            else:
                data['output_subjects'] = [s.subject_uid for s in output_subjects]

        if app_managers is not None:
            data['app_managers'] = app_managers

        resp = connection._post("/1/applications", json=data)

        return CogniacApplication(connection, resp.json())

    ##
    #  get
    ##
    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def get(cls,
            connection,
            application_id):
        """
        return an existing CogniacApplication

        connnection (CogniacConnection):     Authenticated CogniacConnection object
        application_id (String):             The application_id of the Cogniac application to return
        """
        resp = connection._get("/1/applications/%s" % application_id)
        return CogniacApplication(connection, resp.json())

    ##
    #  get_all
    ##
    @classmethod
    def get_all(cls, connection):
        """
        return CogniacApplications for all applications belonging to the currently authenticated tenant

        connnection (CogniacConnection):     Authenticated CogniacConnection object
        """
        resp = connection._get('/1/tenants/%s/applications' % connection.tenant.tenant_id)
        apps = resp.json()['data']
        return [CogniacApplication(connection, appd) for appd in apps]

    ##
    #  __init__
    ##
    def __init__(self, connection, application_dict):
        """
        create a CogniacApplication

        This is not normally called directly by users, instead use:
        CogniacConnection.create_application() or
        CogniacApplication.create()
        """
        self._cc = connection
        self._app_keys = application_dict.keys()
        self.__parse_app_dict__(application_dict)

    def __parse_app_dict__(self, application_dict):
        for k, v in application_dict.items():

            if k == "app_type_config":
                app_type_config = CogniacApplication._CogniacAppTypeConfig(app=self, app_type_config_dict=v)
                super(CogniacApplication, self).__setattr__("app_type_config", app_type_config)
            else:
                super(CogniacApplication, self).__setattr__(k, v)

    ##
    #  delete
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def delete(self):
        """
        Delete the application.
        This will delete existing models but will not delete associated subjects or media.
        """
        self._cc._delete("/1/applications/%s" % self.application_id)

        for k in self._app_keys:
            delattr(self, k)

        self._app_keys = None
        self.connection = None

    def __post_update__(self, data):
        resp = self._cc._post("/1/applications/%s" % self.application_id, json=data)
        self.__parse_app_dict__(resp.json())

    def __setattr__(self, name, value):
        if name in ['app_type_config']:
            raise AttributeError("Please modify {} by accessing its own attributes.".format(name))

        if name in ['application_id', 'created_at', 'created_by', 'modified_at', 'modified_by']:
            raise AttributeError("%s is immutable" % name)

        if name in ['name', 'description', 'active', 'input_subjects', 'output_subjects', 'app_managers',
                    'detection_post_urls', 'detection_thresholds', 'custom_fields', 'app_type_config',
                    'edgeflow_upload_policies', 'override_upstream_detection_filter', 'feedback_resample_ratio',
                    'reviewers', 'inference_execution_policies', 'primary_release_metric', 'secondary_evaluation_metrics']:
            data = {name: value}
            self.__post_update__(data)
            return

        super(CogniacApplication, self).__setattr__(name, value)

    def __str__(self):
        return "%s (%s)" % (self.name, self.application_id)

    def __repr__(self):
        return self.__str__()

    def add_output_subject(self, subject):
        """
        Add a the specified subject the the Application's outputs.

        subject (CogniacSubject):   the subject to add
        """
        self.output_subjects = self.output_subjects + [subject.subject_uid]

    def add_input_subject(self, subject):
        """
        Add a the specified subject the the Application's inputs.

        subject (CogniacSubject):   the subject to add
        """
        self.input_subjects = self.input_subjects + [subject.subject_uid]

    ##
    #  pending_feedback
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def pending_feedback(self):
        """
        Return the integer number of feedback requests pending for this application.
        This is useful for controlling the flow of images input into the system to avoid creating too many backlogged feedback requests.
        """
        resp = self._cc._get("/1/applications/%s/feedback/pending" % self.application_id)
        return resp.json()['pending']

    ##
    #  get_feedback
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def get_feedback(self, limit=10):
        """
        returns a list of  up to {limit} feedback request messages for the application

        limit (Int):   Maximum number of feedback request messages to return
        """
        resp = self._cc._get("/1/applications/%s/feedback?limit=%d" % (self.application_id, limit))
        return resp.json()

    ##
    #  post_feedback
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def post_feedback(self, media_id, subjects):
        """
        Provides feedback to the application for a given subject-media assocation; returns None.

        media_id (String):             Media ID of the media to provide feedback on
        subjects (list of dicts):      Subject-media association dictionaries of the form:

            subject_uid:               Subject UID
            result (str):              One of 'True', 'False', 'Sidelined'
            app_data_type (String):    (Optional) Type of extra app-specific data for certain app types
            app_data (Object):         (Optional) Additional, app-specific, subject-media association data

        """
        # add media_id to each subject-media association dict
        # TODO: deprecate this
        for s in subjects:
            s['media_id'] = media_id

        feedback_response = {'media_id': media_id,
                             'subjects': subjects}

        self._cc._post("/1/applications/%s/feedback" % self.application_id, json=feedback_response)

    ##
    #  list of models released
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def models(self, start=None, end=None, limit=None, reverse=True):
        """
        return a list of models
        """
        # build the search args
        # perform only one search at a time
        args = []
        if start is not None:
            args.append("start=%f" % start)
        if end is not None:
            args.append("end=%f" % end)

        if reverse:
            args.append('reverse=True')
        if limit:
            assert(limit > 0)
            args.append('limit=%d' % min(limit, 100))  # api support max limit of 100

        url = "/1/applications/%s/models?" % self.application_id
        url += "&".join(args)

        @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
        def get_next(url):
            resp = self._cc._get(url)
            return resp.json()

        # due to one model_id can have multiple runtimes, we want to return them together
        # the limit is applied on number of model_ids, not model runtimes
        model_ids = set()
        previous_model_id = None
        while url:
            resp = get_next(url)
            for det in resp['data']:
                model_id = det['model_id']
                model_ids.add(model_id)
                yield det

                if limit and len(model_ids) >= limit:
                    if previous_model_id is not None and model_id != previous_model_id:
                        return
                previous_model_id = model_id
            url = resp['paging'].get('next')
    ##
    #  model_name
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def model_name(self):
        """
        return the modelname for the current best model for this application
        """
        resp = self._cc._get("/1/applications/%s/ccp" % self.application_id)
        resp = resp.json()

        url = resp['best_model_ccp_url']
        modelname = url.split('/')[-1]
        return modelname

    ##
    #  download_model
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def download_model(self, model_id=None):
        """
        Download the current active model for this application to a file in the current working directory and
        return the local filename which will be the same as the model name.
        """
        if model_id is None:
            resp = self._cc._get("/1/applications/%s/ccp" % self.application_id)
            resp = resp.json()
            url = resp['best_model_ccp_url']
            modelname = url.split('/')[-1]
        else:
            modelname = model_id.split('/')[-1]

        resp = self._cc._get("/1/applications/%s/ccppkg" % self.application_id, json={"ccp_filename": modelname})

        fp = open(modelname, "wb")
        fp.write(resp.content)
        fp.close()
        return modelname

    ##
    #  detections
    ##
    def detections(self, start=None, end=None, reverse=True, probability_lower=None, probability_upper=None,  limit=None, consensus_none=False, only_user=False, only_model=False, abridged_media=True):
        """
        Yield application output assertions (model predictions and/or user feedback) sorted by timestamp.


        start (float)          filter by last update timestamp > start (seconds since epoch)
        end (float)            filter by last update timestamp < end   (seconds since epoch)
        reverse (bool)         reverse the sorting order: sort high to low
        probability_lower:     filter by probability > probability_lower
        probability_upper:     filter by probability < probability_upper
        limit (int)            yield maximum of limit results
        consensus_none (bool): only return items that have not reached consensus.
                               This is useful for alternate feedback interfaces where it is
                               undesirable to display items that have already reached consensus
        only_user (bool):      If True, only return feedback assertions from users, not model predictions
        only_model (bool);     If True, only return model prediction assertions, not feedback assertions from users
        abridged_media (bool)  return full media items if False (slower), otherwise return just media_id's for each media_item


        Returns (yield) association dictionary with the following fields:

        media(dict):  { system media dictionary}

        focus(dict):  { system focus context (box, segment, etc) dictionary}

        updated_at(float):  epoch timstamp of model prediction or user feedback

        detections : list of dictionaries as follows:
             detection_id:     internal detection_id
             user_id:          a user_id if this was from a user
             model_id:         a model_id if this was from an app w/model
             uncal_prob:       the raw uncalibrated user or model confidence (if any)
             timestamp:        the time of this detection
             prev_prob:        subject-media probability before this detection (if any)
             probability:      the resulting probability after this detection
             app_data_type     Optional type of extra app-specific data
             app_data          Optional extra app-specific data

        """
        args = []
        if start is not None:
            args.append("start=%f" % start)
        if end is not None:
            args.append("end=%f" % end)
        if probability_lower is not None:
            args.append("probability_lower=%f" % probability_lower)
        if probability_upper is not None:
            args.append("probability_upper=%f" % probability_upper)
        if reverse:
            args.append('reverse=True')
        if limit:
            assert(limit > 0)
            args.append('limit=%d' % min(limit, 100))  # api support max limit of 100
        if consensus_none:
            args.append("consensus_none=True")
        if only_user:
            args.append("only_user=True")
        if only_model:
            args.append("only_model=True")
        if abridged_media:
            args.append('abridged_media=True')

        url = "/1/applications/%s/detections?" % self.application_id
        url += "&".join(args)

        @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
        def get_next(url):
            resp = self._cc._get(url)
            return resp.json()

        count = 0
        while url:
            resp = get_next(url)
            for det in resp['data']:
                yield det
                count += 1
                if limit and count == limit:
                    return
            url = resp['paging'].get('next')

    def accumulate_usage(self, start, end):
        """
        return single cummulative app usage record for start to end epoch times
        """
        url = "/1/usage/app/%s?start=%d&end=%d&accumulate=True" % (self.application_id, start, end)
        resp = self._cc._get(url)
        data = resp.json()['data']
        return data[0] if len(data) else None

    def usage(self, start, end):
        """
        yield sparse app usage records in order between start and end epoch times
        """
        url = "/1/usage/app/%s?start=%d&end=%d" % (self.application_id, start, end)

        @retry(stop_max_attempt_number=8, wait_exponential_multiplier=5, retry_on_exception=server_error)
        def get_next(url):
            resp = self._cc._get(url)
            return resp.json()

        while url:
            resp = get_next(url)
            for record in resp['data']:
                yield record
            url = resp['paging'].get('next')

    class _CogniacAppTypeConfig(object):
        def __init__(self, app, app_type_config_dict):
            super(CogniacApplication._CogniacAppTypeConfig, self).__setattr__(
                "_app_type_config_keys", app_type_config_dict.keys())
            self._app = app

            for k, v in app_type_config_dict.items():
                super(CogniacApplication._CogniacAppTypeConfig, self).__setattr__(k, v)

        def __get_app_type_config_dict__(self):
            d = {}

            for k, v in self.__dict__.items():
                if k in self._app_type_config_keys:
                    d[k] = v

            return d

        def __setattr__(self, name, value):
            if name in self._app_type_config_keys:
                d = self.__get_app_type_config_dict__()
                d[name] = value

                data = {"app_type_config": d}
                self._app.__post_update__(data)
            else:
                super(CogniacApplication._CogniacAppTypeConfig, self).__setattr__(name, value)

        def __repr__(self):
            d = self.__get_app_type_config_dict__()
            return repr(d)



"""
CogniacMedia Object Client

Copyright (C) 2016 Cogniac Corporation
"""

from hashlib import md5
from retrying import retry
from .common import *
from os import stat, fstat, path, SEEK_END
import platform
from time import time

platform_system = platform.system()

immutable_keys = ['frame', 'video', 'media_id', 'size', 'network_camera_id', 'original_url', 'image_width', 'filename', 'original_landing_url', 'uploaded_by_user', 'media_timestamp', 'media_url', 'status', 'hash', 'external_media_id', 'author_profile_url', 'media_src', 'parent_media_id',  'media_resize_urls', 'license', 'tenant_id', 'created_at', 'author', 'public', 'image_height', 'media_format', 'title', 'domain_unit']

mutable_keys =['set_assignment', 'force_set', 'meta_tags', "custom_data"]


def file_creation_time(path_to_file):
    """
    Try to get the date that a file was created, falling back to when it was
    last modified if that isn't possible.
    See http://stackoverflow.com/a/39501288/1709587 for explanation.
    """
    if platform_system == 'Windows':
        return path.getctime(path_to_file)
    else:
        fstat = stat(path_to_file)
        try:
            return fstat.st_birthtime
        except AttributeError:
            # We're probably on Linux. No easy way to get creation dates here,
            # so we'll settle for when its content was last modified.
            return fstat.st_mtime

##
#  CogniacMedia
##
class CogniacMedia(object):
    """
    CogniacMedia objects contain metadata for media files that has been input into the Cogniac System.
    New CogniacMedia can be created by specifying a local filename containing a still image or video.
    Existing CogniacMedia can be retrieved by media_id.
    """

    @classmethod
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def get(cls, connection, media_id):
        """
        return a CogniacMedia object for an existing media item

        connnection (CogniacConnection):     Authenticated CogniacConnection object
        media_id (String):                   The media_id of the Cogniac Media item to return
        """
        resp = connection._get("/1/media/%s" % media_id)

        return CogniacMedia(connection, resp.json())

    @classmethod
    def search(cls, connection, md5=None, filename=None, external_media_id=None, domain_unit=None, limit=None):
        """
        Search for a CogniacMedia item in current tenant by md5, filename, or external_media_id.

        connnection (CogniacConnection):     Authenticated CogniacConnection object
        md5 (String):                        MD5 of media item
        filename (String):                   Original filename of media item
        external_media_id (String):          Customer specified unique ID

        Only one of md5, filename, or external_media_id must be specified

        returns list of CogniacMedia objects that match the query parameters
        """

        if md5 is not None:
            assert((filename is None) & (external_media_id is None) & (domain_unit is None))
            query = "md5=%s" % md5
        elif filename is not None:
            assert((md5 is None) & (external_media_id is None) & (domain_unit is None))
            query = "filename=%s" % filename
        elif external_media_id is not None:
            assert((md5 is None) & (filename is None) & (domain_unit is None))
            query = "external_media_id=%s" % external_media_id
        else:
            assert((md5 is None) & (filename is None) & (external_media_id is None))
            query = "domain_unit=%s" % domain_unit

        def get_next(last_key=None):
            query_limit = limit if limit and limit < 100 else 100
            url = "/1/media/all/search?%s&limit=%s" % (query, query_limit)
            if last_key:
                url += "&last_key=%s" % (last_key)
            resp = connection._get(url)
            return resp.json()

        matches = []
        last_key = None
        while True:
            resp = get_next(last_key=last_key)
            last_key = resp.get('last_key')
            data = [CogniacMedia(connection, m) for m in resp['data']]
            for datum in data:
                matches.append(datum)
                if limit and len(matches) == limit:
                    return matches
            if not last_key:
                break

        return matches

    def __init__(self, connection, media_dict):
        """
        create a CogniacMedia

        This is not normally called directly by users, instead use:
        CogniacConnection.create_media() or
        CogniacMedia.create()
        """
        self._cc = connection
        self._media_keys = media_dict.keys()
        for k, v in media_dict.items():
            super(CogniacMedia, self).__setattr__(k, v)

    def __str__(self):
        return "%s" % self.media_id

    def __setattr__(self, name, value):
        if name in immutable_keys:
            raise AttributeError("%s is immutable" % name)
        if name in mutable_keys:
            data = {name: value}
            resp = self._cc._post("/1/media/%s" % self.media_id, json=data)
            for k, v in resp.json().items():
                super(CogniacMedia, self).__setattr__(k, v)
            return
        super(CogniacMedia, self).__setattr__(name, value)

    @classmethod
    def create(cls,
               connection,
               filename,
               meta_tags=None,
               force_set=None,
               set_assignment=None,
               external_media_id=None,
               original_url=None,
               original_landing_url=None,
               license=None,
               author_profile_url=None,
               author=None,
               title=None,
               media_timestamp=None,
               domain_unit=None,
               trigger_id=None,
               sequence_ix=None,
               custom_data=None,
               fp=None):
        """
        Create a new CogniacMedia object and upload the media to the Cogniac System.

        connnection (CogniacConnection):  Authenticated CogniacConnection object
        filename (str):                   Local filename or http/s URL of image or video media file
        meta_tags ([str]):                Optional list of arbitrary strings to associate with the media
        force_set (str):                  [DEPRECATED] Optionally force the media into the 'training', 'validation' or 'test' sets
        set_assignment (str):             Optionally force the media into the 'training', 'validation' or 'test' sets
        external_media_id (str):          Optional arbitrary external id for this media
        original_url(str):                Optional source url for this media
        original_landing_url (str):       Optional source landing url for this media
        license (str):                    Optional copyright licensing info for this media
        author_profile_url (str):         Optional media author url
        author (str):                     Optional author name
        title (str):                      Optional media title
        media_timestamp (float):          Optional actual timestamp of media creation/occurance time
        domain_unit (str):                Optional domain id (e.g. serial number) for set assignment grouping.
                                          Media with the same domain_unit will always be assigned to the same
                                          training or validation set. Set this to avoid overfitting when you have
                                          multiple images of the same thing or almost the same thing.
        trigger_id (str):                 Unique trigger identifier leading to a media sequence containing this media
        sequence_ix (str):                The index of this media within a triggered sequence
        custom_data (str):                Opaque user-specified data associated with this media; limited to 32KB
        fp (file):                        A '.read()'-supporting file-like object from which to acquire
                                          the media instead of reading the media from the specified filename.
        """

        args = dict()
        if meta_tags is not None:
            args['meta_tags'] = meta_tags
        if force_set is not None:
            args['force_set'] = force_set
        if set_assignment is not None:
            args['set_assignment'] = set_assignment
        if external_media_id is not None:
            args['external_media_id'] = external_media_id
        if original_url is not None:
            args['original_url'] = original_url
        if original_landing_url is not None:
            args['original_landing_url'] = original_landing_url
        if license is not None:
            args['license'] = license
        if author_profile_url is not None:
            args['author_profile_url'] = author_profile_url
        if author is not None:
            args['author'] = author
        if title is not None:
            args['title'] = title
        if media_timestamp is not None:
            args['media_timestamp'] = media_timestamp
        if domain_unit is not None:
            args['domain_unit'] = domain_unit
        if trigger_id is not None:
            args['trigger_id'] = trigger_id
        if sequence_ix is not None:
            args['sequence_ix'] = sequence_ix
        if custom_data is not None:
            args['custom_data'] = custom_data

        if filename.startswith('http'):
            args['source_url'] = filename
        else:
            if 'media_timestamp' not in args:
                if fp is None:
                    # set the unspecified media timestamp to the earliest file time we have
                    args['media_timestamp'] = file_creation_time(filename)
                else:
                    args['media_timestamp'] = time()

            if fp is None:
                fsize = stat(filename).st_size
                # use the multipart interface for large files
            else:
                fp.seek(0, SEEK_END) # most reliable way of finding file size.
                fsize = fp.tell()
                fp.seek(0)

            if fsize > 12 * 1024 * 1024:
                return CogniacMedia._create_multipart(connection, filename, fp, fsize, args)

        @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
        def upload():
            if filename.startswith('http'):
                files = None
            elif fp is not None:
                fp.seek(0)  # for the retry win
                files = {filename: fp}
            else:
                files = {filename: open(filename, 'rb')}
            resp = connection._post("/1/media", data=args, files=files)
            return resp

        resp = upload()
        return CogniacMedia(connection, resp.json())

    @classmethod
    def _create_multipart(cls, connection, filename, mfp, filesize, args):
        """
        upload via the multipart api
        """
        def md5_hexdigest(fp):
            """
            return the md5 hexdigest of a potentially very large file
            """
            md = md5()
            while True:
                block = fp.read(8*1024*1024)
                if not block:
                    return md.hexdigest()
                md.update(block)

        if mfp is None:
            mfp = open(filename, 'rb')
        else:
            mfp.seek(0)

        md5hash = md5_hexdigest(mfp)

        @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
        def post_data(data):
            resp = connection._post("/1/media/resumable", json=data)
            return resp.json()

        data = {'upload_phase': 'start',
                'file_size':    filesize,
                'filename':     filename,
                'md5':          md5hash}

        rdata = post_data(data)
        upload_session_id = rdata['upload_session_id']
        chunk_size = rdata['chunk_size']

        @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
        def upload_chunk(chunk, chunk_no, upload_session_id):

            data = {'upload_phase':        'transfer',
                    'upload_session_id':   upload_session_id,
                    'video_file_chunk_no': chunk_no}

            files = {'file': chunk}
            resp = connection._post("/1/media/resumable", data=data, files=files)
            return resp.json()

        idx = 1
        mfp.seek(0)
        while True:
            chunk = mfp.read(chunk_size)
            if not chunk:
                break
            upload_chunk(chunk, idx, upload_session_id)
            idx += 1

        # perform the final finihs phase post with the user args
        data = {'upload_phase':      'finish',
                'upload_session_id': upload_session_id}
        data.update(args)  # add user arg
        rdata = post_data(data)
        return CogniacMedia(connection, rdata)

    def delete(self):
        """
        delete the media object
        """
        self._cc._delete("/1/media/%s" % self.media_id)
        self.__dict__.clear()

    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def download(self, filep=None, timeout=60):
        """
        Download the media file and return as a string or write to a file.

        filep:   open file object to store downloaded media
                 If filep is None: return the media as a string

        timeout: timeout in seconds
        """
        url = self.media_url

        stream = False
        if filep is not None:
            stream = True  # user requests output to file so stream potentially large results
        resp = self._cc._get(url, stream=stream, timeout=timeout)
        raise_errors(resp)

        if filep is None:  # return response all at once
            return resp.content
        else:
            filep.seek(0)  # in case of retries

        # write response out to file
        for chunk in resp.iter_content(chunk_size=512 * 1024):
            if chunk:  # filter out keep-alive new chunks
                filep.write(chunk)
        filep.close()

    ##
    #  detections
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def detections(self, wait_capture_id=None):
        """
        return a list of detection dictionaries as follows for the specified media_id and this subject
        wait_capture_id (str):    capture_id from previous subject associate_media call that had enable_wait_result=True
                                  This operation with block and return all applications detections resulting from capture_id
                                  when available.

        Returns:

        detection_id:     internal detection_id
        user_id:          a user_id if this was from a user
        model_id:         a model_id if this was from an app w/model
        app_id:           an app_id if this was from an app
        uncal_prob:       the raw uncalibrated user or model confidence (if any)
        timestamp:        the time of this detection
        prev_prob:        subject-media probability before this detection (if any)
        probability:      the resulting probability after this detection
        app_data_type     Optional type of extra app-specific data
        app_data          Optional extra app-specific data
        """

        url = "/1/media/%s/detections" % (self.media_id)

        if wait_capture_id is not None:
            url += "?wait_capture_id=%s" % wait_capture_id

        resp = self._cc._get(url)
        return resp.json()['detections']

    ##
    #  subjects
    ##
    @retry(stop_max_attempt_number=8, wait_exponential_multiplier=500, retry_on_exception=server_error)
    def subjects(self):
        """
        return a list of subjects as follows for the specified media_id:

        {"media":       {partial or full media item},

         "subject:"     {

        media_id        the media_id
        subject_uid		the subject_uid
        probability		current assessment of the probability in [0,1] that the subject_uid is associated with the media_id
                        1 = definitely associated
                        0 = definitely NOT associated
                        0.5 ~= uncertain association
        timestamp		time of last update
        app_data_type	optional app data type if applicable
        app_data        optional app data if applicable
        consensus		'True', 'False', 'Sidelined', or None
                        'True' if there is consensus that the subject is associated with the media
                            (Media will be used as a positive training example of the subject)
                        'False' if there is consensus that the subject is not associated w/the media
                            (Media will be used as a negative training example of the subject.)
                         None if if there is not enough evidence to reach consensus
                         'Sidelined' if there is consensus that the media should not be used 
                            for training applications dependent on the subject.
                         Some application types only support 'True' or None.
                         }
        """
        resp = self._cc._get("/1/media/%s/subjects" % (self.media_id))
        return resp.json()['data']
